from __future__ import annotations

from typing import TYPE_CHECKING, overload

from django.db import models
from django.db.models import CASCADE, PROTECT

from lamindb.base import deprecated
from lamindb.base.fields import (
    BooleanField,
    CharField,
    DateTimeField,
    ForeignKey,
    TextField,
)
from lamindb.errors import FieldValidationError

from ..base.ids import base62_8
from .can_curate import CanCurate
from .feature import Feature
from .has_parents import HasParents, _query_ancestors_of_fk
from .run import Run, TracksRun, TracksUpdates, User, current_user_id
from .sqlrecord import BaseSQLRecord, IsLink, SQLRecord, _get_record_kwargs
from .transform import Transform

if TYPE_CHECKING:
    from datetime import datetime

    from .artifact import Artifact
    from .collection import Collection
    from .project import Project
    from .query_set import SQLRecordList


class ULabel(SQLRecord, HasParents, CanCurate, TracksRun, TracksUpdates):
    """Universal labels.

    For new labels, see `Record` instead. Existing labels and code will continue to work
    but be migrated to the Record registry.

    Args:
        name: `str` A name.
        description: `str | None = None` A description.
        reference: `str | None = None` For instance, an external ID or a URL.
        reference_type: `str | None = None` For instance, `"url"`.

    See Also:
        :meth:`~lamindb.Feature`
            Dimensions of measurement for artifacts & collections.
        :attr:`~lamindb.Artifact.features`
            Feature manager for an artifact.

    Examples:

        Create a ulabel:

        >>> train_split = ln.ULabel(name="train").save()

        Organize ulabels in a hierarchy:

        >>> split_type = ln.ULabel(name="Split", is_type=True).save()
        >>> train_split = ln.ULabel(name="train", type="split_type").save()

        Label an artifact:

        >>> artifact.ulabels.add(ulabel)

        Query an artifact by ulabel:

        >>> ln.Artifact.filter(ulabels=train_split).to_dataframe()
    """

    class Meta(SQLRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        app_label = "lamindb"

    _name_field: str = "name"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = CharField(
        editable=False, unique=True, db_index=True, max_length=8, default=base62_8
    )
    """A universal random id, valid across DB instances."""
    name: str = CharField(max_length=150, db_index=True)
    """Name or title of ulabel."""
    type: ULabel | None = ForeignKey("self", PROTECT, null=True, related_name="ulabels")
    """Type of ulabel, e.g., `"donor"`, `"split"`, etc.

    Allows to group ulabels by type, e.g., all donors, all split ulabels, etc.
    """
    ulabels: ULabel
    """ULabels of this type (can only be non-empty if `is_type` is `True`)."""
    is_type: bool = BooleanField(default=False, db_index=True, null=True)
    """Distinguish types from instances of the type.

    For example, a ulabel "Project" would be a type, and the actual projects "Project 1", "Project 2", would be records of that `type`.
    """
    description: str | None = TextField(null=True)
    """A description."""
    reference: str | None = CharField(max_length=255, db_index=True, null=True)
    """A simple reference like URL or external ID."""
    reference_type: str | None = CharField(max_length=25, db_index=True, null=True)
    """Type of simple reference."""
    parents: ULabel = models.ManyToManyField(
        "self", symmetrical=False, related_name="children"
    )
    """Parent entities of this ulabel.

    For advanced use cases, you can build an ontology under a given `type`.

    Say, if you modeled `CellType` as a `ULabel`, you would introduce a type `CellType` and model the hiearchy of cell types under it.
    """
    children: ULabel
    """Child entities of this ulabel.

    Reverse accessor for parents.
    """
    transforms: Transform
    """Linked transforms."""
    runs: Run
    """Linked runs."""
    artifacts: Artifact
    """Linked artifacts."""
    collections: Collection
    """Linked collections."""
    projects: Project
    """Linked projects."""

    @overload
    def __init__(
        self,
        name: str,
        type: ULabel | None = None,
        is_type: bool = False,
        description: str | None = None,
        reference: str | None = None,
        reference_type: str | None = None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        if len(args) == len(self._meta.concrete_fields):
            super().__init__(*args, **kwargs)
            return None
        if len(args) > 0:
            raise ValueError("Only one non-keyword arg allowed")
        name: str = kwargs.pop("name", None)
        type: str | None = kwargs.pop("type", None)
        is_type: bool = kwargs.pop("is_type", False)
        description: str | None = kwargs.pop("description", None)
        reference: str | None = kwargs.pop("reference", None)
        reference_type: str | None = kwargs.pop("reference_type", None)
        branch = kwargs.pop("branch", None)
        branch_id = kwargs.pop("branch_id", 1)
        space = kwargs.pop("space", None)
        space_id = kwargs.pop("space_id", 1)
        _skip_validation = kwargs.pop("_skip_validation", False)
        _aux = kwargs.pop("_aux", None)
        if len(kwargs) > 0:
            valid_keywords = ", ".join([val[0] for val in _get_record_kwargs(ULabel)])
            raise FieldValidationError(
                f"Only {valid_keywords} are valid keyword arguments"
            )
        super().__init__(
            name=name,
            type=type,
            is_type=is_type,
            description=description,
            reference=reference,
            reference_type=reference_type,
            branch=branch,
            branch_id=branch_id,
            space=space,
            space_id=space_id,
            _skip_validation=_skip_validation,
            _aux=_aux,
        )

    def query_types(self) -> SQLRecordList:
        """Query types of a ulabel recursively.

        While `.type` retrieves the direct type, this method
        retrieves all ancestors of that `type`.
        """
        return _query_ancestors_of_fk(self, "type")  # type: ignore

    @property
    @deprecated("ulabels")
    def records(self) -> list[ULabel]:
        """Return all instances of this type."""
        return self.ulabels


class ArtifactULabel(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = ForeignKey("Artifact", CASCADE, related_name="links_ulabel")
    ulabel: ULabel = ForeignKey(ULabel, PROTECT, related_name="links_artifact")
    feature: Feature | None = ForeignKey(
        Feature, PROTECT, null=True, related_name="links_artifactulabel", default=None
    )
    label_ref_is_name: bool | None = BooleanField(null=True)
    feature_ref_is_name: bool | None = BooleanField(null=True)

    class Meta:
        # can have the same label linked to the same artifact if the feature is
        # different
        app_label = "lamindb"
        unique_together = ("artifact", "ulabel", "feature")


class TransformULabel(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    transform: Transform = ForeignKey(Transform, CASCADE, related_name="links_ulabel")
    ulabel: ULabel = ForeignKey(ULabel, PROTECT, related_name="links_transform")

    class Meta:
        app_label = "lamindb"
        unique_together = ("transform", "ulabel")


class RunULabel(BaseSQLRecord, IsLink):
    id: int = models.BigAutoField(primary_key=True)
    run: Run = ForeignKey(Run, CASCADE, related_name="links_ulabel")
    ulabel: ULabel = ForeignKey(ULabel, PROTECT, related_name="links_run")
    created_at: datetime = DateTimeField(
        editable=False, db_default=models.functions.Now(), db_index=True
    )
    """Time of creation of record."""
    created_by: User = ForeignKey(
        "lamindb.User", PROTECT, default=current_user_id, related_name="+"
    )
    """Creator of record."""

    class Meta:
        app_label = "lamindb"
        unique_together = ("run", "ulabel")


class CollectionULabel(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    collection: Collection = ForeignKey(
        "Collection", CASCADE, related_name="links_ulabel"
    )
    ulabel: ULabel = ForeignKey(ULabel, PROTECT, related_name="links_collection")
    feature: Feature | None = ForeignKey(
        Feature, PROTECT, null=True, related_name="links_collectionulabel", default=None
    )
    label_ref_is_name: bool | None = BooleanField(null=True)
    feature_ref_is_name: bool | None = BooleanField(null=True)

    class Meta:
        app_label = "lamindb"
        unique_together = ("collection", "ulabel")



# Generated by Django 5.2 on 2025-07-04 20:43

import django.db.models.deletion
from django.db import migrations

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0111_remove_record__sort_order"),
    ]

    operations = [
        migrations.AlterField(
            model_name="recordartifact",
            name="feature",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_recordartifact",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="recordjson",
            name="feature",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_recordjson",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="recordrecord",
            name="feature",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_recordrecord",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="recordrun",
            name="feature",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_recordrun",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="recordulabel",
            name="feature",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_recordulabel",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="recordperson",
            name="feature",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_recordperson",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="recordproject",
            name="feature",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_recordproject",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="recordreference",
            name="feature",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_recordreference",
                to="lamindb.feature",
            ),
        ),
        migrations.AlterField(
            model_name="recordartifact",
            name="value",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_in_record",
                to="lamindb.artifact",
            ),
        ),
    ]



# Generated by Django 5.2 on 2025-05-11 18:42

import django.db.models.deletion
import django.db.models.functions.datetime
from django.db import migrations, models

import lamindb.base.fields
import lamindb.base.users
import lamindb.models.sqlrecord


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0094_writeloglock_writelogmigrationstate_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="RunFeatureValue",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "featurevalue",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_run",
                        to="lamindb.featurevalue",
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_featurevalue",
                        to="lamindb.run",
                    ),
                ),
            ],
            options={
                "unique_together": {("run", "featurevalue")},
            },
            bases=(models.Model, lamindb.models.sqlrecord.IsLink),
        ),
        migrations.AddField(
            model_name="run",
            name="_feature_values",
            field=models.ManyToManyField(
                related_name="runs",
                through="lamindb.RunFeatureValue",
                to="lamindb.featurevalue",
            ),
        ),
        migrations.RemoveField(
            model_name="flextable",
            name="created_by",
        ),
        migrations.RemoveField(
            model_name="flextable",
            name="projects",
        ),
        migrations.RemoveField(
            model_name="flextable",
            name="schema",
        ),
        migrations.RemoveField(
            model_name="flextable",
            name="space",
        ),
        migrations.RemoveField(
            model_name="flextable",
            name="type",
        ),
        migrations.RemoveField(
            model_name="flextable",
            name="ulabels",
        ),
        migrations.RemoveField(
            model_name="flextabledata",
            name="space",
        ),
        migrations.RemoveField(
            model_name="flextabledata",
            name="value_artifact",
        ),
        migrations.RemoveField(
            model_name="flextabledata",
            name="value_collection",
        ),
        migrations.RemoveField(
            model_name="flextabledata",
            name="value_person",
        ),
        migrations.RemoveField(
            model_name="flextabledata",
            name="value_project",
        ),
        migrations.RemoveField(
            model_name="flextabledata",
            name="value_ulabel",
        ),
        migrations.RemoveField(
            model_name="rundata",
            name="space",
        ),
        migrations.RemoveField(
            model_name="rundata",
            name="value_artifact",
        ),
        migrations.RemoveField(
            model_name="rundata",
            name="value_collection",
        ),
        migrations.RemoveField(
            model_name="rundata",
            name="value_person",
        ),
        migrations.RemoveField(
            model_name="rundata",
            name="value_project",
        ),
        migrations.RemoveField(
            model_name="rundata",
            name="value_ulabel",
        ),
        migrations.DeleteModel(
            name="FlexTable",
        ),
        migrations.DeleteModel(
            name="FlexTableData",
        ),
        migrations.DeleteModel(
            name="RunData",
        ),
    ]



import shutil
from pathlib import Path

import pytest


@pytest.fixture(scope="function")
def clean_soma_files(request):
    path = request.param if hasattr(request, "param") else "small_dataset.tiledbsoma"
    if Path(path).exists():
        shutil.rmtree(path)

    yield path

    if Path(path).exists():
        shutil.rmtree(path)



from __future__ import annotations

from typing import TYPE_CHECKING, overload

from django.core.validators import RegexValidator
from django.db import models
from django.db.models import CASCADE, PROTECT

from lamindb.base.fields import (
    BigIntegerField,
    BooleanField,
    CharField,
    DateField,
    DateTimeField,
    ForeignKey,
    TextField,
    URLField,
)
from lamindb.base.users import current_user_id

from ..base.ids import base62_12
from .artifact import Artifact
from .can_curate import CanCurate
from .collection import Collection
from .feature import Feature
from .record import Record
from .run import Run, TracksRun, TracksUpdates, User
from .schema import Schema
from .sqlrecord import BaseSQLRecord, IsLink, SQLRecord, ValidateFields
from .transform import Transform
from .ulabel import ULabel

if TYPE_CHECKING:
    from datetime import date as DateType
    from datetime import datetime

    from .block import ProjectBlock


class Reference(SQLRecord, CanCurate, TracksRun, TracksUpdates, ValidateFields):
    """References such as internal studies, papers, documents, or URLs.

    Example:
        >>> reference = Reference(
        ...     name="A Paper Title",
        ...     abbr="APT",
        ...     url="https://doi.org/10.1000/xyz123",
        ...     pubmed_id=12345678,
        ...     doi="10.1000/xyz123",
        ...     description="Good paper.",
        ...     text="Some text I want to be searchable.",
        ...     date=date(2023, 11, 21),
        ... ).save()
    """

    class Meta(SQLRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        app_label = "lamindb"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = CharField(
        editable=False, unique=True, max_length=12, db_index=True, default=base62_12
    )
    """Universal id, valid across DB instances."""
    name: str = CharField(db_index=True)
    """Title or name of the reference document."""
    description: str | None = TextField(null=True)
    """A description."""
    type: Reference | None = ForeignKey(
        "self", PROTECT, null=True, related_name="references"
    )
    """Type of reference (e.g., 'Study', 'Paper', 'Preprint').

    Allows to group reference by type, e.g., internal studies vs. all papers etc.
    """
    references: Reference
    """References of this type (can only be non-empty if `is_type` is `True`)."""
    is_type: bool = BooleanField(default=False, db_index=True, null=True)
    """Distinguish types from instances of the type."""
    abbr: str | None = CharField(
        max_length=32,
        db_index=True,
        null=True,
    )
    """An abbreviation for the reference."""
    url: str | None = URLField(null=True, db_index=True)
    """URL linking to the reference."""
    pubmed_id: int | None = BigIntegerField(null=True, db_index=True)
    """A PudMmed ID."""
    doi: str | None = CharField(
        null=True,
        db_index=True,
        validators=[
            RegexValidator(
                regex=r"^(?:https?://(?:dx\.)?doi\.org/|doi:|DOI:)?10\.\d+/.*$",
                message="Must be a DOI (e.g., 10.1000/xyz123 or https://doi.org/10.1000/xyz123)",
            )
        ],
    )
    """Digital Object Identifier (DOI) for the reference."""
    text: str | None = TextField(null=True, db_index=True)
    """Abstract or full text of the reference to make it searchable."""
    date: DateType | None = DateField(null=True, default=None)
    """Date of creation or publication of the reference."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact, through="ArtifactReference", related_name="references"
    )
    """Annotated artifacts."""
    transforms: Artifact = models.ManyToManyField(
        Transform, through="TransformReference", related_name="references"
    )
    """Annotated transforms."""
    collections: Artifact = models.ManyToManyField(
        Collection, through="CollectionReference", related_name="references"
    )
    """Annotated collections."""
    linked_in_records: Record = models.ManyToManyField(
        Record, through="RecordReference", related_name="linked_references"
    )
    """Linked in records."""
    records: Record = models.ManyToManyField(
        Record, through="ReferenceRecord", related_name="references"
    )
    """Annotated records."""
    projects: Project
    """Projects that annotate this reference."""

    @overload
    def __init__(
        self,
        name: str,
        type: Reference | None = None,
        is_type: bool = False,
        abbr: str | None = None,
        url: str | None = None,
        pubmed_id: int | None = None,
        doi: str | None = None,
        description: str | None = None,
        text: str | None = None,
        date: DateType | None = None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Project(SQLRecord, CanCurate, TracksRun, TracksUpdates, ValidateFields):
    """Projects to label artifacts, transforms, and runs.

    Example::

        project = Project(
           name="My Project Name",
           abbr="MPN",
           url="https://example.com/my_project",
        ).save()
        artifact.projects.add(project)  # <-- labels the artifact with the project
        ln.track(project=project)       # <-- automtically labels entities during the run

    """

    class Meta(SQLRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        app_label = "lamindb"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = CharField(
        editable=False, unique=True, max_length=12, db_index=True, default=base62_12
    )
    """Universal id, valid across DB instances."""
    name: str = CharField(db_index=True)
    """Title or name of the Project."""
    description: str | None = TextField(null=True)
    """A description."""
    type: Project | None = ForeignKey(
        "self", PROTECT, null=True, related_name="projects"
    )
    """Type of project (e.g., 'Program', 'Project', 'GithubIssue', 'Task')."""
    projects: Project
    """Projects of this type (can only be non-empty if `is_type` is `True`)."""
    is_type: bool = BooleanField(default=False, db_index=True, null=True)
    """Distinguish types from instances of the type."""
    abbr: str | None = CharField(max_length=32, db_index=True, null=True)
    """An abbreviation."""
    url: str | None = URLField(max_length=255, null=True, default=None)
    """A URL."""
    start_date: DateType | None = DateField(null=True, default=None)
    """Date of start of the project."""
    end_date: DateType | None = DateField(null=True, default=None)
    """Date of start of the project."""
    parents: Project = models.ManyToManyField(
        "self", symmetrical=False, related_name="children"
    )
    """Parent projects, the super-projects owning this project."""
    children: Project
    """Child projects, the sub-projects owned by this project.

    Reverse accessor for `.parents`.
    """
    predecessors: Project = models.ManyToManyField(
        "self", symmetrical=False, related_name="successors"
    )
    """The preceding projects required by this project."""
    successors: Project
    """The succeeding projects requiring this project.

    Reverse accessor for `.predecessors`.
    """
    artifacts: Artifact = models.ManyToManyField(
        Artifact, through="ArtifactProject", related_name="projects"
    )
    """Annotated artifacts."""
    transforms: Transform = models.ManyToManyField(
        Transform, through="TransformProject", related_name="projects"
    )
    """Annotated transforms."""
    runs: Run = models.ManyToManyField(
        Run, through="RunProject", related_name="projects"
    )
    """Annotated runs."""
    ulabels: ULabel = models.ManyToManyField(
        ULabel, through="ULabelProject", related_name="projects"
    )
    """Annotated ulabels."""
    features: ULabel = models.ManyToManyField(
        Feature, through="FeatureProject", related_name="projects"
    )
    """Annotated features."""
    schemas: ULabel = models.ManyToManyField(
        Schema, through="SchemaProject", related_name="projects"
    )
    """Annotated schemas."""
    linked_in_records: Record = models.ManyToManyField(
        Record, through="RecordProject", related_name="linked_projects"
    )
    """Linked in records."""
    records: Record = models.ManyToManyField(
        Record, through="ProjectRecord", related_name="projects"
    )
    """Annotated records."""
    collections: Collection = models.ManyToManyField(
        Collection, through="CollectionProject", related_name="projects"
    )
    """Annotated collections."""
    references: Reference = models.ManyToManyField("Reference", related_name="projects")
    """Annotated references."""
    _status_code: int = models.SmallIntegerField(default=0, db_index=True)
    """Status code."""
    blocks: ProjectBlock
    """Blocks that annotate this project."""

    @overload
    def __init__(
        self,
        name: str,
        type: Project | None = None,
        is_type: bool = False,
        abbr: str | None = None,
        url: str | None = None,
        start_date: DateType | None = None,
        end_date: DateType | None = None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ArtifactProject(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = ForeignKey(Artifact, CASCADE, related_name="links_project")
    project: Project = ForeignKey(Project, PROTECT, related_name="links_artifact")
    feature: Feature | None = ForeignKey(
        Feature,
        PROTECT,
        null=True,
        default=None,
        related_name="links_artifactproject",
    )
    label_ref_is_name: bool | None = BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = BooleanField(null=True, default=None)

    class Meta:
        app_label = "lamindb"
        # can have the same label linked to the same artifact if the feature is different
        unique_together = ("artifact", "project", "feature")


class RunProject(BaseSQLRecord, IsLink):
    id: int = models.BigAutoField(primary_key=True)
    run: Run = ForeignKey(Run, CASCADE, related_name="links_project")
    project: Project = ForeignKey(Project, PROTECT, related_name="links_run")
    created_at: datetime = DateTimeField(
        editable=False, db_default=models.functions.Now(), db_index=True
    )
    """Time of creation of record."""
    created_by: User = ForeignKey(
        "lamindb.User",
        PROTECT,
        editable=False,
        default=current_user_id,
        related_name="+",
    )
    """Creator of record."""

    class Meta:
        app_label = "lamindb"
        unique_together = ("run", "project")


class TransformProject(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    transform: Transform = ForeignKey(Transform, CASCADE, related_name="links_project")
    project: Project = ForeignKey(Project, PROTECT, related_name="links_transform")

    class Meta:
        app_label = "lamindb"
        unique_together = ("transform", "project")


class CollectionProject(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    collection: Collection = ForeignKey(
        Collection, CASCADE, related_name="links_project"
    )
    project: Project = ForeignKey(Project, PROTECT, related_name="links_collection")

    class Meta:
        app_label = "lamindb"
        unique_together = ("collection", "project")


class ULabelProject(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    ulabel: ULabel = ForeignKey(ULabel, CASCADE, related_name="links_project")
    project: Project = ForeignKey(Project, PROTECT, related_name="links_ulabel")

    class Meta:
        app_label = "lamindb"
        unique_together = ("ulabel", "project")


class FeatureProject(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    feature: Feature = ForeignKey(Feature, CASCADE, related_name="links_project")
    project: Project = ForeignKey(Project, PROTECT, related_name="links_feature")

    class Meta:
        app_label = "lamindb"
        unique_together = ("feature", "project")


class SchemaProject(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    schema: Schema = ForeignKey(Schema, CASCADE, related_name="links_project")
    project: Project = ForeignKey(Project, PROTECT, related_name="links_schema")

    class Meta:
        app_label = "lamindb"
        unique_together = ("schema", "project")


# for annotation of records with references, RecordReference is for storing reference values
class ReferenceRecord(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    reference: Reference = ForeignKey(Reference, PROTECT, related_name="links_record")
    feature: Feature | None = ForeignKey(
        Feature,
        PROTECT,
        null=True,
        default=None,
        related_name="links_referencerecord",
    )
    record: Record = ForeignKey(Record, CASCADE, related_name="links_reference")

    class Meta:
        app_label = "lamindb"
        unique_together = ("reference", "feature", "record")


class RecordReference(BaseSQLRecord, IsLink):
    id: int = models.BigAutoField(primary_key=True)
    record: Record = ForeignKey(Record, CASCADE, related_name="values_reference")
    feature: Feature = ForeignKey(
        Feature, PROTECT, related_name="links_recordreference"
    )
    value: Reference = ForeignKey(Reference, PROTECT, related_name="links_in_record")

    class Meta:
        app_label = "lamindb"
        unique_together = ("record", "feature", "value")


# for annotation of records with projects, RecordProject is for storing project values
class ProjectRecord(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    project: Project = ForeignKey(Project, PROTECT, related_name="links_record")
    feature: Feature | None = ForeignKey(
        Feature,
        PROTECT,
        null=True,
        default=None,
        related_name="links_projectrecord",
    )
    record: Record = ForeignKey(Record, CASCADE, related_name="links_project")

    class Meta:
        app_label = "lamindb"
        unique_together = ("project", "feature", "record")


class RecordProject(BaseSQLRecord, IsLink):
    id: int = models.BigAutoField(primary_key=True)
    record: Record = ForeignKey(Record, CASCADE, related_name="values_project")
    feature: Feature = ForeignKey(Feature, PROTECT, related_name="links_recordproject")
    value: Project = ForeignKey(Project, PROTECT, related_name="links_in_record")

    class Meta:
        app_label = "lamindb"
        unique_together = ("record", "feature", "value")


class ArtifactReference(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = ForeignKey(Artifact, CASCADE, related_name="links_reference")
    reference: Reference = ForeignKey(Reference, PROTECT, related_name="links_artifact")
    feature: Feature | None = ForeignKey(
        Feature,
        PROTECT,
        null=True,
        default=None,
        related_name="links_artifactreference",
    )
    label_ref_is_name: bool | None = BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = BooleanField(null=True, default=None)

    class Meta:
        app_label = "lamindb"
        unique_together = ("artifact", "reference", "feature")


class TransformReference(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    transform: Transform = ForeignKey(
        Transform, CASCADE, related_name="links_reference"
    )
    reference: Reference = ForeignKey(
        Reference, PROTECT, related_name="links_transform"
    )

    class Meta:
        app_label = "lamindb"
        unique_together = ("transform", "reference")


class CollectionReference(BaseSQLRecord, IsLink, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    collection: Collection = ForeignKey(
        Collection, CASCADE, related_name="links_reference"
    )
    reference: Reference = ForeignKey(
        Reference, PROTECT, related_name="links_collection"
    )

    class Meta:
        app_label = "lamindb"
        unique_together = ("collection", "reference")



# Generated by Django 5.2 on 2025-10-13 07:42

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0133_artifactuser_artifact_users"),
    ]

    operations = [
        migrations.AddField(
            model_name="run",
            name="params",
            field=models.JSONField(null=True),
        ),
    ]



"""Examples and utilities for Weights & Biases.

.. autosummary::
   :toctree: .

   save_wandb_features
"""

import lamindb as ln


def save_wandb_features():
    """Saves all Weights & Biases project and run related features.

    Saves the following features:

    - wandb_run_id
    - wandb_run_name
    - wandb_run_entity
    - wandb_project
    - wandb_state
    - wandb_url
    - wandb_tags
    - wandb_group
    - wandb_job_type
    - timestamp
    - runtime
    """
    wandb_type = ln.Feature(name="Weights & Biases", is_type=True).save()
    ln.Feature(name="wandb_run_id", dtype=str, type=wandb_type).save()
    ln.Feature(name="wandb_run_name", dtype=str, type=wandb_type).save()
    ln.Feature(name="wandb_run_entity", dtype=str, type=wandb_type).save()
    ln.Feature(name="wandb_project", dtype=str, type=wandb_type).save()
    ln.Feature(name="wandb_state", dtype=str, type=wandb_type).save()
    ln.Feature(name="wandb_url", dtype=str, type=wandb_type).save()
    ln.Feature(name="wandb_tags", dtype=str, type=wandb_type).save()
    ln.Feature(name="wandb_group", dtype=str, type=wandb_type).save()
    ln.Feature(name="wandb_job_type", dtype=str, type=wandb_type).save()
    ln.Feature(name="wandb_timestamp", dtype=float, type=wandb_type).save()
    ln.Feature(name="wandb_runtime", dtype=float, type=wandb_type).save()



import concurrent.futures

import lamindb as ln


# we need this test both in the core and the storage/cloud tests
# because the internal logic that retrieves information about other instances
# depends on whether the current instance is managed on the hub
def test_reference_storage_location(ccaplog):
    ln.Artifact("s3://lamindata/iris_studies/study0_raw_images")
    assert ln.Storage.get(root="s3://lamindata").instance_uid == "4XIuR0tvaiXM"
    assert (
        "referenced read-only storage location at s3://lamindata, is managed by instance with uid 4XIuR0tvaiXM"
        in ccaplog.text
    )


def test_create_storage_locations_parallel():
    root: str = "nonregistered_storage"

    def create_storage() -> str:
        ln.Storage(root=root).save()  # type: ignore
        return root

    n_parallel = 3
    with concurrent.futures.ThreadPoolExecutor(max_workers=n_parallel) as executor:
        futures = [executor.submit(create_storage) for i in range(n_parallel)]
        _ = [future.result() for future in concurrent.futures.as_completed(futures)]

    storage = ln.Storage.get(root__endswith=root)
    storage.delete()



import lamindb as ln

schema = ln.Schema(
    name="Mini immuno schema",
    features=[
        ln.Feature.get(name="perturbation"),
        ln.Feature.get(name="cell_type_by_model"),
        ln.Feature.get(name="assay_oid"),
        ln.Feature.get(name="donor"),
        ln.Feature.get(name="concentration"),
        ln.Feature.get(name="treatment_time_h"),
    ],
    flexible=True,  # _additional_ columns in a dataframe are validated & annotated
).save()



import bionty as bt
import lamindb as ln
import pytest
from lamindb.examples.cellxgene._cellxgene import CELLxGENESchemaVersions


@pytest.fixture
def cxg_schema_factory():
    def create_schema(version: CELLxGENESchemaVersions, **kwargs):
        ln.examples.cellxgene.save_cellxgene_defaults()
        schema = ln.examples.cellxgene.create_cellxgene_schema(
            schema_version=version, **kwargs
        )
        return schema

    yield create_schema

    # Cleanup after all tests
    ln.models.SchemaComponent.filter().delete(permanent=True)
    ln.Schema.filter().delete(permanent=True)
    ln.Feature.filter().delete(permanent=True)
    ln.ULabel.filter(type__isnull=False).delete(permanent=True)
    for entity in [
        bt.Disease,
        bt.Ethnicity,
        bt.DevelopmentalStage,
        bt.Phenotype,
        bt.CellType,
        ln.ULabel,
    ]:
        entity.filter().delete(permanent=True)


def test_cxg_curator_5(cxg_schema_factory):
    cxg_schema = cxg_schema_factory("5.2.0", field_types=["name", "ontology_id"])

    # test invalid var index and typo in obs column
    adata = ln.examples.datasets.small_dataset3_cellxgene(
        with_obs_defaults=True, with_obs_typo=True, with_var_typo=True
    )
    curator = ln.curators.AnnDataCurator(adata, cxg_schema)
    # Ensure that default values for Features are set
    curator.slots["obs"].standardize()
    with pytest.raises(ln.errors.ValidationError) as e:
        curator.validate()
        assert (
            "ValidationError: 1 term not validated in feature 'index' in slot 'var': 'invalid_ensembl_id'"
            in str(e)
        )

    # subset adata to remove invalid var index
    adata = adata[
        :, ~adata.var.index.isin(curator.slots["var"].cat.non_validated["index"])
    ].copy()
    curator = ln.curators.AnnDataCurator(adata, cxg_schema)
    with pytest.raises(ln.errors.ValidationError) as e:
        curator.validate()
        assert (
            "ValidationError: 1 term not validated in feature 'tissue_ontology_term_id' in slot 'obs': 'UBERON:0002048XXX'"
            in str(e)
        )
    # fix typo in obs column
    adata.obs["tissue_ontology_term_id"] = adata.obs["tissue_ontology_term_id"].replace(
        {"UBERON:0002048XXX": "UBERON:0002048"}
    )
    artifact = curator.save_artifact(key="examples/dataset-curated-against-cxg-5.h5ad")

    # test missing obs columns
    adata = ln.examples.datasets.small_dataset3_cellxgene(with_obs_defaults=False)
    adata = adata[:, ~adata.var.index.isin({"invalid_ensembl_id"})].copy()
    curator = ln.curators.AnnDataCurator(adata, cxg_schema)
    with pytest.raises(ln.errors.ValidationError) as e:
        curator.validate()
    expected_missing = [
        "assay",
        "cell_type",
        "development_stage",
        "disease",
        "self_reported_ethnicity",
        "assay_ontology_term_id",
        "cell_type_ontology_term_id",
        "development_stage_ontology_term_id",
        "self_reported_ethnicity_ontology_term_id",
        "tissue_ontology_term_id",
        "organism_ontology_term_id",
        "donor_id",
    ]
    assert all(col in str(e.value) for col in expected_missing)

    # Clean up
    artifact.delete(permanent=True)


def test_cxg_curator_6_spatial(cxg_schema_factory):
    """Tests organism (in `uns` as of 6.x) and spatial slot validation of CELLxGENE 6.x."""
    cxg_schema = cxg_schema_factory(
        "6.0.0", spatial_library_id="library_123", field_types="ontology_id"
    )

    adata = ln.examples.datasets.small_dataset3_cellxgene(
        with_obs_defaults=True, with_uns_organism=True, with_uns_spatial=True
    )
    # delete a necessary component from uns["spatial"]
    del adata.uns["spatial"]["is_single"]

    curator = ln.curators.AnnDataCurator(adata, cxg_schema)

    with pytest.raises(ln.errors.ValidationError) as e:
        curator.validate()
    assert "column 'is_single' not in dataframe." in str(e.value)

    adata.uns["spatial"]["is_single"] = True
    curator = ln.curators.AnnDataCurator(adata, cxg_schema)
    curator.validate()


def test_invalid_field_type():
    with pytest.raises(ValueError) as e:
        ln.examples.cellxgene.create_cellxgene_schema(
            schema_version="5.3.0", field_types=["ensembl_gene_ids"]
        )
    assert "Invalid field_types" in str(e.value)


def test_invalid_schema_Version():
    with pytest.raises(ValueError) as e:
        ln.examples.cellxgene.create_cellxgene_schema(schema_version="200.0.0")
    assert "Invalid schema_version" in str(e.value)



# Generated by Django 5.2 on 2025-06-30 12:42

import django.db.models.deletion
from django.db import migrations

import lamindb.base.fields


def migrate_sheets_to_records(apps, schema_editor):
    """Migrate Sheet records to Record table as type records."""
    with schema_editor.connection.cursor() as cursor:
        # Insert sheets as records with is_type=True
        cursor.execute("""
            INSERT INTO lamindb_record (uid, name, description, schema_id, is_type, created_by_id, created_at, updated_at, run_id)
            SELECT uid, name, description, schema_id, TRUE, created_by_id, created_at, updated_at, run_id
            FROM lamindb_sheet;
        """)

        # Update records that were linked to sheets to link to the new record types
        cursor.execute("""
            UPDATE lamindb_record
            SET type_id = (
                SELECT r.id
                FROM lamindb_record r
                JOIN lamindb_sheet s ON r.uid = s.uid
                WHERE s.id = lamindb_record.sheet_id
            )
            WHERE sheet_id IS NOT NULL;
        """)


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0106_transfer_data_migration"),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name="record",
            name="unique_name",
        ),
        migrations.AddField(
            model_name="record",
            name="schema",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="records",
                to="lamindb.schema",
            ),
        ),
        migrations.AlterField(
            model_name="record",
            name="is_type",
            field=lamindb.base.fields.BooleanField(
                blank=True, db_index=True, default=False
            ),
        ),
        migrations.RunPython(
            migrate_sheets_to_records,
        ),
        migrations.AddField(
            model_name="record",
            name="_sort_order",
            field=django.db.models.FloatField(null=True, default=None),
        ),
    ]



"""Curator utilities.

.. autosummary::
   :toctree: .

   Curator
   SlotsCurator
   ComponentCurator
   CatVector
   CatLookup
   DataFrameCatManager

"""

from __future__ import annotations

import copy
import re
from typing import TYPE_CHECKING, Any, Callable

import lamindb_setup as ln_setup
import numpy as np
import pandas as pd
import pandera.pandas as pandera
from lamin_utils import colors, logger
from lamindb_setup.core._docs import doc_args
from lamindb_setup.core.upath import LocalPathClasses

from lamindb.base.dtypes import check_dtype
from lamindb.base.types import FieldAttr  # noqa
from lamindb.models import (
    Artifact,
    Feature,
    Run,
    Schema,
    SQLRecord,
)
from lamindb.models._from_values import _format_values
from lamindb.models.artifact import (
    data_is_scversedatastructure,
    data_is_soma_experiment,
)
from lamindb.models.feature import (
    parse_cat_dtype,
    parse_dtype,
    parse_filter_string,
    resolve_relation_filters,
)

from ..errors import InvalidArgument, ValidationError

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Any

    from anndata import AnnData
    from mudata import MuData
    from spatialdata import SpatialData
    from tiledbsoma._experiment import Experiment as SOMAExperiment

    from lamindb.core.types import ScverseDataStructures
    from lamindb.models.query_set import SQLRecordList


def strip_ansi_codes(text):
    # This pattern matches ANSI escape sequences
    ansi_pattern = re.compile(r"\x1b\[[0-9;]*m")
    return ansi_pattern.sub("", text)


class CatLookup:
    """Lookup categories from the reference instance.

    Args:
        categoricals: A dictionary of categorical fields to lookup.
        slots: A dictionary of slot fields to lookup.
        public: Whether to lookup from the public instance. Defaults to False.

    Example::

        curator = ln.curators.DataFrameCurator(...)
        curator.cat.lookup()["cell_type"].alveolar_type_1_fibroblast_cell

    """

    def __init__(
        self,
        categoricals: list[Feature] | dict[str, FieldAttr],
        slots: dict[str, FieldAttr] = None,
        public: bool = False,
        sources: dict[str, SQLRecord] | None = None,
    ) -> None:
        slots = slots or {}
        if isinstance(categoricals, list):
            categoricals = {
                feature.name: parse_dtype(feature.dtype)[0]["field"]
                for feature in categoricals
            }
        self._categoricals = {**categoricals, **slots}
        self._public = public
        self._sources = sources

    def __getattr__(self, name):
        if name in self._categoricals:
            registry = self._categoricals[name].field.model
            if self._public and hasattr(registry, "public"):
                return registry.public(source=self._sources.get(name)).lookup()
            else:
                return registry.lookup()
        raise AttributeError(
            f'"{self.__class__.__name__}" object has no attribute "{name}"'
        )

    def __getitem__(self, name):
        if name in self._categoricals:
            registry = self._categoricals[name].field.model
            if self._public and hasattr(registry, "public"):
                return registry.public(source=self._sources.get(name)).lookup()
            else:
                return registry.lookup()
        raise AttributeError(
            f'"{self.__class__.__name__}" object has no attribute "{name}"'
        )

    def __repr__(self) -> str:
        if len(self._categoricals) > 0:
            getattr_keys = "\n ".join(
                [f".{key}" for key in self._categoricals if key.isidentifier()]
            )
            getitem_keys = "\n ".join(
                [str([key]) for key in self._categoricals if not key.isidentifier()]
            )
            ref = "public" if self._public else "registries"
            return (
                f"Lookup objects from the {colors.italic(ref)}:\n "
                f"{colors.green(getattr_keys)}\n "
                f"{colors.green(getitem_keys)}\n"
                'Example:\n    → categories = curator.lookup()["cell_type"]\n'
                "    → categories.alveolar_type_1_fibroblast_cell\n\n"
                "To look up public ontologies, use .lookup(public=True)"
            )
        else:  # pragma: no cover
            return colors.warning("No fields are found!")


CAT_MANAGER_DOCSTRING = """Manage categoricals by updating registries."""


SLOTS_DOCSTRING = """Access sub curators by slot."""

SLOTS_DETAILS_DOCSTRING = """Uses **slots** to specify which component contains which schema. Slots are keys that identify where features are stored within composite data structures."""

VALIDATE_DOCSTRING = """Validate dataset against Schema.

Raises:
    lamindb.errors.ValidationError: If validation fails.
"""

SAVE_ARTIFACT_DOCSTRING = """Save an annotated artifact.

Args:
    key: A path-like key to reference artifact in default storage, e.g., `"myfolder/myfile.fcs"`. Artifacts with the same key form a version family.
    description: A description.
    revises: Previous version of the artifact. Is an alternative way to passing `key` to trigger a new version.
    run: The run that creates the artifact.

Returns:
    A saved artifact record.
"""


class Curator:
    """Curator base class.

    A `Curator` object makes it easy to validate, standardize & annotate datasets.

    See:
        - :class:`~lamindb.curators.DataFrameCurator`
        - :class:`~lamindb.curators.AnnDataCurator`
        - :class:`~lamindb.curators.MuDataCurator`
        - :class:`~lamindb.curators.SpatialDataCurator`
        - :class:`~lamindb.curators.TiledbsomaExperimentCurator`
    """

    def __init__(self, dataset: Any, schema: Schema | None = None):
        if not isinstance(schema, Schema):
            raise InvalidArgument("schema argument must be a Schema record.")

        if schema.pk is None:
            raise ValueError(
                "Schema must be saved before curation. Please save it using '.save()'."
            )

        self._artifact: Artifact = None  # pass the dataset as an artifact
        self._dataset: Any = dataset  # pass the dataset as a UPathStr or data object
        if isinstance(self._dataset, Artifact):
            self._artifact = self._dataset
            if self._artifact.otype in {
                "DataFrame",
                "AnnData",
                "MuData",
                "SpatialData",
            }:
                # Open remote AnnData Artifacts
                if not isinstance(self._artifact.path, LocalPathClasses):
                    if self._artifact.otype in {
                        "AnnData",
                    }:
                        try:
                            self._dataset = self._dataset.open(mode="r")
                        # open can raise various errors. Fall back to loading into memory if open fails
                        except Exception as e:
                            logger.warning(
                                f"Unable to open remote AnnData Artifact: {e}. Falling back to loading into memory."
                            )
                            self._dataset = self._dataset.load(is_run_input=False)
                else:
                    self._dataset = self._dataset.load(is_run_input=False)
        self._schema: Schema | None = schema
        self._is_validated: bool = False

    @doc_args(VALIDATE_DOCSTRING)
    def validate(self) -> bool | str:
        """{}"""  # noqa: D415
        pass  # pragma: no cover

    @doc_args(SAVE_ARTIFACT_DOCSTRING)
    def save_artifact(
        self,
        *,
        key: str | None = None,
        description: str | None = None,
        revises: Artifact | None = None,
        run: Run | None = None,
    ) -> Artifact:
        """{}"""  # noqa: D415
        # Note that this docstring has to be consistent with the Artifact()
        # constructor signature
        pass  # pragma: no cover

    def __repr__(self) -> str:
        from lamin_utils import colors

        if self._schema is not None:
            # Schema might have different attributes
            if hasattr(self._schema, "name") and self._schema.name:
                schema_str = colors.italic(self._schema.name)
            elif hasattr(self._schema, "uid"):
                schema_str = colors.italic(f"uid={self._schema.uid}")
            elif hasattr(self._schema, "id"):
                schema_str = colors.italic(f"id={self._schema.id}")
            else:
                schema_str = colors.italic("unnamed")

            # Add schema type info if available
            if hasattr(self._schema, "otype") and self._schema.otype:
                schema_str += f" ({self._schema.otype})"
        else:
            schema_str = colors.warning("None")

        status_str = ""
        if self._is_validated:
            status_str = f", {colors.green('validated')}"
        else:
            status_str = f", {colors.yellow('unvalidated')}"

        cls_name = colors.green(self.__class__.__name__)

        # Get additional info based on curator type
        extra_info = ""
        if hasattr(self, "_slots") and self._slots:
            # For SlotsCurator and its subclasses
            slots_count = len(self._slots)
            if slots_count > 0:
                slot_names = list(self._slots.keys())
                if len(slot_names) <= 3:
                    extra_info = f", slots: {slot_names}"
                else:
                    extra_info = f", slots: [{', '.join(slot_names[:3])}... +{len(slot_names) - 3} more]"
        elif (
            cls_name == "DataFrameCurator"
            and hasattr(self, "cat")
            and hasattr(self.cat, "_categoricals")
        ):
            # For DataFrameCurator
            cat_count = len(getattr(self.cat, "_categoricals", []))
            if cat_count > 0:
                extra_info = f", categorical_features={cat_count}"

        artifact_info = ""
        if self._artifact is not None:
            artifact_info = f", artifact: {colors.italic(self._artifact.uid)}"

        return (
            f"{cls_name}{artifact_info}(Schema: {schema_str}{extra_info}{status_str})"
        )


@doc_args(SLOTS_DETAILS_DOCSTRING)
class SlotsCurator(Curator):
    """Curator for a dataset with slots.

    {}

    Args:
        dataset: The dataset to validate & annotate.
        schema: A :class:`~lamindb.Schema` object that defines the validation constraints.
    """

    def __init__(
        self,
        dataset: Artifact | ScverseDataStructures | SOMAExperiment,
        schema: Schema,
    ) -> None:
        super().__init__(dataset=dataset, schema=schema)
        self._slots: dict[str, ComponentCurator] = {}

        # used for multimodal data structures (not AnnData)
        # in form of {table/modality_key: var_field}
        self._var_fields: dict[str, FieldAttr] = {}
        # in form of {table/modality_key: categoricals}
        self._cat_vectors: dict[str, dict[str, CatVector]] = {}

    @property
    @doc_args(SLOTS_DOCSTRING)
    def slots(self) -> dict[str, ComponentCurator]:
        """{}"""  # noqa: D415
        return self._slots

    @doc_args(VALIDATE_DOCSTRING)
    def validate(self) -> None:
        """{}"""  # noqa: D415
        for slot, curator in self._slots.items():
            logger.debug(f"validating slot {slot} ...")
            curator.validate()
        # set _is_validated to True as no slot raised an error
        self._is_validated = True

    @doc_args(SAVE_ARTIFACT_DOCSTRING)
    def save_artifact(
        self,
        *,
        key: str | None = None,
        description: str | None = None,
        revises: Artifact | None = None,
        run: Run | None = None,
    ) -> Artifact:
        """{}"""  # noqa: D415
        if not self._is_validated:
            self.validate()

        if self._artifact is None:
            type_mapping = [
                (
                    lambda dataset: isinstance(dataset, pd.DataFrame),
                    Artifact.from_dataframe,
                ),
                (
                    lambda dataset: data_is_scversedatastructure(dataset, "AnnData"),
                    Artifact.from_anndata,
                ),
                (
                    lambda dataset: data_is_scversedatastructure(dataset, "MuData"),
                    Artifact.from_mudata,
                ),
                (
                    lambda dataset: data_is_scversedatastructure(
                        dataset, "SpatialData"
                    ),
                    Artifact.from_spatialdata,
                ),
                (data_is_soma_experiment, Artifact.from_tiledbsoma),
            ]

            for type_check, af_constructor in type_mapping:
                if type_check(self._dataset):
                    self._artifact = af_constructor(  # type: ignore
                        self._dataset,
                        key=key,
                        description=description,
                        revises=revises,
                        run=run,
                    )
                    break

        cat_vectors = {}
        for curator in self._slots.values():
            for key, cat_vector in curator.cat._cat_vectors.items():
                cat_vectors[key] = cat_vector

        self._artifact.schema = self._schema
        self._artifact.save()
        return annotate_artifact(  # type: ignore
            self._artifact,
            curator=self,
            cat_vectors=cat_vectors,
        )


def convert_dict_to_dataframe_for_validation(d: dict, schema: Schema) -> pd.DataFrame:
    """Convert a dictionary to a DataFrame for validation against a schema."""
    df = pd.DataFrame([d])
    for feature in schema.members:
        if feature.dtype.startswith("cat"):
            if feature.name in df.columns:
                df[feature.name] = pd.Categorical(df[feature.name])
    return df


# This is also currently used as DictCurator by flattening dictionaries into wide DataFrames.
# Such an approach was never intended and there is room for a DictCurator in the future.
# For more context, read https://laminlabs.slack.com/archives/C07DB677JF6/p1753994077716099 and
# https://www.notion.so/laminlabs/Add-a-DictCurator-2422aeaa55e180b9a513f91d13970836
class ComponentCurator(Curator):
    """Curator for `DataFrame`.

    Provides all key functionality to validate Pandas DataFrames.
    This class is not user facing unlike :class:`~lamindb.curators.DataFrameCurator` which extends this
    class with functionality to validate the `attrs` slot.

    Args:
        dataset: The DataFrame-like object to validate & annotate.
        schema: A :class:`~lamindb.Schema` object that defines the validation constraints.
        slot: Indicate the slot in a composite curator for a composite data structure.
    """

    def __init__(
        self,
        dataset: pd.DataFrame | Artifact,
        schema: Schema,
        slot: str | None = None,
    ) -> None:
        super().__init__(dataset=dataset, schema=schema)

        categoricals = []
        features = []
        feature_ids: set[int] = set()

        if schema.flexible:
            features += Feature.filter(name__in=self._dataset.keys()).to_list()
            feature_ids = {feature.id for feature in features}

        if schema.n > 0:
            if schema._index_feature_uid is not None:
                schema_features = [
                    feature
                    for feature in schema.members.to_list()
                    if feature.uid != schema._index_feature_uid  # type: ignore
                ]
            else:
                schema_features = schema.members.to_list()  # type: ignore
            if feature_ids:
                features.extend(
                    feature
                    for feature in schema_features
                    if feature.id not in feature_ids  # type: ignore
                )
            else:
                features.extend(schema_features)
        else:
            assert schema.itype is not None  # noqa: S101

        pandera_columns = {}
        if features or schema._index_feature_uid is not None:
            # populate features
            if schema.minimal_set:
                optional_feature_uids = set(schema.optionals.get_uids())
            for feature in features:
                if schema.minimal_set:
                    required = feature.uid not in optional_feature_uids
                else:
                    required = False
                # series.dtype is "object" if the column has lists types, e.g. [["a", "b"], ["a"], ["b"]]
                if feature.dtype in {
                    "int",
                    "float",
                    "num",
                    "path",
                } or feature.dtype.startswith("list"):
                    if isinstance(self._dataset, pd.DataFrame):
                        dtype = (
                            self._dataset[feature.name].dtype
                            if feature.name in self._dataset.keys()
                            else None
                        )
                    else:
                        dtype = None
                    pandera_columns[feature.name] = pandera.Column(
                        dtype=None,
                        checks=pandera.Check(
                            check_dtype(feature.dtype),
                            element_wise=False,
                            error=f"Column '{feature.name}' failed dtype check for '{feature.dtype}': got {dtype}",
                        ),
                        nullable=feature.nullable,
                        coerce=feature.coerce_dtype,
                        required=required,
                    )
                else:
                    pandera_dtype = (
                        feature.dtype
                        if not feature.dtype.startswith("cat")
                        else "category"
                    )
                    pandera_columns[feature.name] = pandera.Column(
                        pandera_dtype,
                        nullable=feature.nullable,
                        coerce=feature.coerce_dtype,
                        required=required,
                    )
                if feature.dtype.startswith("cat") or feature.dtype.startswith(
                    "list[cat["
                ):
                    # validate categoricals if the column is required or if the column is present
                    # but exclude the index feature from column categoricals
                    if (required or feature.name in self._dataset.keys()) and (
                        schema._index_feature_uid is None
                        or feature.uid != schema._index_feature_uid
                    ):
                        categoricals.append(feature)
            # in almost no case, an index should have a pandas.CategoricalDtype in a DataFrame
            # so, we're typing it as `str` here
            if schema.index is not None:
                index = pandera.Index(
                    schema.index.dtype
                    if not schema.index.dtype.startswith("cat")
                    else str
                )
            else:
                index = None

            self._pandera_schema = pandera.DataFrameSchema(
                pandera_columns,
                coerce=schema.coerce_dtype,
                strict=schema.maximal_set,
                ordered=schema.ordered_set,
                index=index,
            )
        # in the DataFrameCatManager, we use the
        # actual columns of the dataset, not the pandera columns
        # the pandera columns might have additional optional columns
        if schema.itype == "Composite":
            columns_field = Feature.name
        else:
            columns_field = parse_cat_dtype(schema.itype, is_itype=True)["field"]
        self._cat_manager = DataFrameCatManager(
            self._dataset,
            columns_field=columns_field,
            categoricals=categoricals,
            index=schema.index,
            slot=slot,
            maximal_set=schema.maximal_set,
        )

    @property
    @doc_args(CAT_MANAGER_DOCSTRING)
    def cat(self) -> DataFrameCatManager:
        """{}"""  # noqa: D415
        return self._cat_manager

    def standardize(self) -> None:
        """Standardize the dataset.

        - Adds missing columns for features
        - Fills missing values for features with default values
        """
        if self._artifact is not None:
            raise RuntimeError(
                "Cannot mutate the dataset when an artifact is passed! Please load the dataset into memory using `dataset.load()` and pass it to a curator."
            )

        for feature in self._schema.members:
            if feature.name not in self._dataset.columns:
                if feature.default_value is not None or feature.nullable:
                    fill_value = (
                        feature.default_value
                        if feature.default_value is not None
                        else pd.NA
                    )
                    if feature.dtype.startswith("cat"):
                        self._dataset[feature.name] = pd.Categorical(
                            [fill_value] * len(self._dataset)
                        )
                    else:
                        self._dataset[feature.name] = fill_value
                    logger.important(
                        f"added column {feature.name} with fill value {fill_value}"
                    )
                else:
                    raise ValidationError(
                        f"Missing column {feature.name} cannot be added because is not nullable and has no default value"
                    )
            else:
                if feature.default_value is not None:
                    if isinstance(
                        self._dataset[feature.name].dtype, pd.CategoricalDtype
                    ):
                        if (
                            feature.default_value
                            not in self._dataset[feature.name].cat.categories
                        ):
                            self._dataset[feature.name] = self._dataset[
                                feature.name
                            ].cat.add_categories(feature.default_value)
                    self._dataset[feature.name] = self._dataset[feature.name].fillna(
                        feature.default_value
                    )

    def _cat_manager_validate(self) -> None:
        self.cat.validate()

        if self.cat._is_validated:
            self._is_validated = True
        else:
            self._is_validated = False
            raise ValidationError(self.cat._validate_category_error_messages)

    @doc_args(VALIDATE_DOCSTRING)
    def validate(self) -> None:
        """{}"""  # noqa: D415
        if self._schema.n > 0:
            try:
                # first validate through pandera
                self._pandera_schema.validate(self._dataset, lazy=True)
                # then validate lamindb categoricals
                self._cat_manager_validate()
            except (pandera.errors.SchemaError, pandera.errors.SchemaErrors) as err:
                self._is_validated = False
                has_dtype_error = "WRONG_DATATYPE" in str(err)
                error_msg = str(err)
                if has_dtype_error:
                    error_msg += "   ▶ Hint: Consider setting 'coerce_dtype=True' to attempt coercing/converting values during validation to the pre-defined dtype."
                raise ValidationError(error_msg) from err
        else:
            self._cat_manager_validate()

    @doc_args(SAVE_ARTIFACT_DOCSTRING)
    def save_artifact(
        self,
        *,
        key: str | None = None,
        description: str | None = None,
        revises: Artifact | None = None,
        run: Run | None = None,
    ) -> Artifact:
        """{}"""  # noqa: D415
        if not self._is_validated:
            self.validate()  # raises ValidationError if doesn't validate
        if self._artifact is None:
            self._artifact = Artifact.from_dataframe(
                self._dataset,
                key=key,
                description=description,
                revises=revises,
                run=run,
                format=".csv" if key is not None and key.endswith(".csv") else None,
            )

        self._artifact.schema = self._schema
        self._artifact.save()
        return annotate_artifact(  # type: ignore
            self._artifact,
            cat_vectors=self.cat._cat_vectors,
        )


class DataFrameCurator(SlotsCurator):
    # the example in the docstring is tested in test_curators_quickstart_example
    """Curator for `DataFrame`.

    Args:
        dataset: The DataFrame-like object to validate & annotate.
        schema: A :class:`~lamindb.Schema` object that defines the validation constraints.
        slot: Indicate the slot in a composite curator for a composite data structure.

    Examples:

        For a simple example using a flexible schema, see :meth:`~lamindb.Artifact.from_dataframe`.

        Here is an example that enforces a minimal set of columns in the dataframe.

        .. literalinclude:: scripts/curate_dataframe_minimal_errors.py
            :language: python

        Under-the-hood, this used the following schema.

        .. literalinclude:: scripts/define_mini_immuno_schema_flexible.py
            :language: python

        Valid features & labels were defined as:

        .. literalinclude:: scripts/define_mini_immuno_features_labels.py
            :language: python

        It is also possible to curate the `attrs` slot.

        .. literalinclude:: scripts/curate_dataframe_attrs.py
            :language: python
    """

    def __init__(
        self,
        dataset: pd.DataFrame | Artifact,
        schema: Schema,
        slot: str | None = None,
    ) -> None:
        super().__init__(dataset=dataset, schema=schema)

        self._atomic_curator = ComponentCurator(
            dataset=dataset,
            schema=schema,
            slot=slot,
        )

        # Handle (nested) attrs
        if slot is None and schema.slots:
            for slot_name, slot_schema in schema.slots.items():
                if slot_name.startswith("attrs"):
                    path_parts = slot_name.split(":")
                    attrs_dict = getattr(self._dataset, "attrs", None)
                    if attrs_dict is not None:
                        if len(path_parts) == 1:
                            data = attrs_dict
                        else:
                            deeper_keys = path_parts[1:]
                            data = _resolve_schema_slot_path(
                                attrs_dict, deeper_keys, slot_name, "attrs"
                            )
                        df = convert_dict_to_dataframe_for_validation(data, slot_schema)
                        self._slots[slot_name] = ComponentCurator(
                            df, slot_schema, slot=slot_name
                        )
                elif slot_name != "__external__":
                    raise ValueError(
                        f"Slot '{slot_name}' is not supported for DataFrameCurator. Must be 'attrs'."
                    )

    @property
    def cat(self) -> DataFrameCatManager:
        """Manage categoricals by updating registries."""
        if hasattr(self, "_atomic_curator"):
            return self._atomic_curator.cat
        raise AttributeError("cat is only available for slots DataFrameCurator")

    def standardize(self) -> None:
        """Standardize the dataset.

        - Adds missing columns for features
        - Fills missing values for features with default values
        """
        if hasattr(self, "_atomic_curator"):
            self._atomic_curator.standardize()
        else:
            for slot_curator in self._slots.values():
                slot_curator.standardize()

    @doc_args(VALIDATE_DOCSTRING)
    def validate(self) -> None:
        """{}."""
        if hasattr(self, "_atomic_curator"):
            self._atomic_curator.validate()
            self._is_validated = self._atomic_curator._is_validated
        if self._schema.itype == "Composite":
            super().validate()

    @doc_args(SAVE_ARTIFACT_DOCSTRING)
    def save_artifact(
        self, *, key=None, description=None, revises=None, run=None
    ) -> Artifact:
        """{}."""
        if not self._is_validated:
            self.validate()

        if self._slots:
            self._slots["columns"] = self._atomic_curator
            try:
                return super().save_artifact(
                    key=key, description=description, revises=revises, run=run
                )
            finally:
                del self._slots["columns"]
        else:
            return self._atomic_curator.save_artifact(
                key=key, description=description, revises=revises, run=run
            )


class ExperimentalDictCurator(DataFrameCurator):
    """Curator for `dict` based on `DataFrameCurator`."""

    def __init__(
        self,
        dataset: dict | Artifact,
        schema: Schema,
        slot: str | None = None,
    ) -> None:
        if not isinstance(dataset, dict) and not isinstance(dataset, Artifact):
            raise InvalidArgument("The dataset must be a dict or dict-like artifact.")
        if isinstance(dataset, Artifact):
            assert dataset.otype == "dict", "Artifact must be of otype 'dict'."  # noqa: S101
            d = dataset.load(is_run_input=False)
        else:
            d = dataset
        df = convert_dict_to_dataframe_for_validation(d, schema)
        super().__init__(df, schema, slot=slot)


def _resolve_schema_slot_path(
    target_dict: dict[str, Any], slot_keys: Iterable[str], slot: str, base_path: str
) -> Any:
    """Resolve a schema slot path by traversing nested dictionary keys.

    Args:
        target_dict: Root dictionary to traverse
        slot_keys: Sequence of keys defining the paths to traverse
        slot_name: Schema slot identifier for error context
        base_path: Base path string for error context

    Returns:
        The value at the resolved path
    """
    current = target_dict

    for key in slot_keys:
        base_path += f"['{key}']"
        try:
            current = current[key]
        except (
            KeyError,
            TypeError,
        ):  # if not a dict, raises TypeError; if a dict and key not found, raises KeyError
            available = (
                list(current.keys())
                if isinstance(current, dict)
                else "none (not a dict)"
            )
            raise InvalidArgument(
                f"Schema slot '{slot}' requires keys {base_path} but key '{key}' "
                f"not found. Available keys at this level: {available}."
            ) from None

    return current


def _handle_dict_slots(
    dataset: ScverseDataStructures, slot: str
) -> tuple[pd.DataFrame | None, str | None, str | None]:
    """Handle dict-based slot paths (uns/attrs standalone or of modalities) for all ScverseCurators.

    Supports two patterns:
        - Direct dict access: "uns", "attrs", "uns:key1:key2", "attrs:key"
        - Modality dict access: "modality:uns"

    Args:
        dataset: The scverse datastructure object
        slot: The slot path string to parse like 'uns:path:to'.

    Returns:
        tuple: (dataframe, modality_key, remaining_slot_path)
            - dataframe: Single-row DataFrame containing the resolved data
            - modality_key: Modality identifier if slot targets modality dict, else None
            - remaining_slot_path: The dict attribute and nested keys as string
    """
    path_parts = slot.split(":")

    # Handle direct dict slots: "uns", "attrs", "uns:key1:key2:..."
    if len(path_parts) >= 1 and path_parts[0] in ["uns", "attrs"]:
        dict_attr = getattr(dataset, path_parts[0], None)
        if dict_attr is not None:
            if len(path_parts) == 1:
                return pd.DataFrame([dict_attr]), None, path_parts[0]

            deeper_keys = path_parts[1:]
            data = _resolve_schema_slot_path(
                dict_attr, deeper_keys, slot, path_parts[0]
            )
            return pd.DataFrame([data]), None, ":".join(path_parts[1:])

    # Handle modality dict slots: "modality:uns", "modality:uns:key1:key2"
    elif len(path_parts) >= 2 and path_parts[1] in ["uns", "attrs"]:
        modality, dict_name = path_parts[0], path_parts[1]
        try:
            modality_dataset = dataset[modality]
            dict_attr = getattr(modality_dataset, dict_name, None)
            if dict_attr is not None:
                if len(path_parts) == 2:
                    return pd.DataFrame([dict_attr]), modality, dict_name

                deeper_keys = path_parts[2:]
                data = _resolve_schema_slot_path(
                    dict_attr, deeper_keys, slot, f"{modality}.{dict_name}"
                )
                return pd.DataFrame([data]), modality, ":".join(path_parts[1:])
        except (KeyError, AttributeError):
            pass
    else:
        raise InvalidArgument(
            f"Invalid dict slot pattern '{slot}'. Expected formats: "
            f"'uns', 'attrs', 'uns:key', 'attrs:key', 'modality:uns'"
        )

    return None, None, None


@doc_args(SLOTS_DETAILS_DOCSTRING)
class AnnDataCurator(SlotsCurator):
    """Curator for `AnnData`.

    {}

    Args:
        dataset: The AnnData-like object to validate & annotate.
        schema: A :class:`~lamindb.Schema` object that defines the validation constraints.

    Examples:

        Curate Ensembl gene IDs and valid features in obs:

        .. literalinclude:: scripts/curate_anndata_flexible.py
            :language: python
            :caption: curate_anndata_flexible.py

        Curate `uns` dictionary:

        .. literalinclude:: scripts/curate_anndata_uns.py
            :language: python
            :caption: curate_anndata_uns.py

    See Also:
        :meth:`~lamindb.Artifact.from_anndata`.
    """

    def __init__(
        self,
        dataset: AnnData | Artifact,
        schema: Schema,
    ) -> None:
        super().__init__(dataset=dataset, schema=schema)
        if not data_is_scversedatastructure(self._dataset, "AnnData"):
            raise InvalidArgument("dataset must be AnnData-like.")
        if schema.otype != "AnnData":
            raise InvalidArgument("Schema otype must be 'AnnData'.")

        for slot, slot_schema in schema.slots.items():
            if slot not in {"var", "var.T", "obs"} and not slot.startswith("uns"):
                raise ValueError(
                    f"AnnDataCurator currently only supports the slots 'var', 'var.T', 'obs', and 'uns', not {slot}"
                )
            if slot.startswith("uns"):
                df, _, _ = _handle_dict_slots(self._dataset, slot)
            elif slot in {"obs", "var", "var.T"}:
                df = (
                    getattr(self._dataset, slot.strip(".T")).T
                    if slot == "var.T"
                    or (
                        slot == "var"
                        and schema.slots["var"].itype not in {None, "Feature"}
                    )
                    else getattr(self._dataset, slot)
                )
            self._slots[slot] = ComponentCurator(df, slot_schema, slot=slot)

            # Handle var index naming for backward compat
            if slot == "var" and schema.slots["var"].itype not in {None, "Feature"}:
                logger.warning(
                    "auto-transposed `var` for backward compat, please indicate transposition in the schema definition by calling out `.T`: slots={'var.T': itype=bt.Gene.ensembl_gene_id}"
                )
                self._slots["var"].cat._cat_vectors["var_index"] = self._slots[
                    "var"
                ].cat._cat_vectors.pop("columns")
                self._slots["var"].cat._cat_vectors["var_index"]._key = "var_index"


def _assign_var_fields_categoricals_multimodal(
    modality: str | None,
    slot_type: str,
    slot: str,
    slot_schema: Schema,
    var_fields: dict[str, FieldAttr],
    cat_vectors: dict[str, dict[str, CatVector]],
    slots: dict[str, ComponentCurator],
) -> None:
    """Assigns var_fields and categoricals for multimodal data curators."""
    if modality is not None:
        var_fields[modality] = None
        cat_vectors[modality] = {}

    if slot_type == "var":
        var_field = parse_cat_dtype(slot_schema.itype, is_itype=True)["field"]
        if modality is None:
            # This should rarely/never be used since tables should have different var fields
            var_fields[slot] = var_field  # pragma: no cover
        else:
            # Note that this is NOT nested since the nested key is always "var"
            var_fields[modality] = var_field
    else:
        obs_fields = slots[slot].cat._cat_vectors
        if modality is None:
            cat_vectors[slot] = obs_fields
        else:
            # Note that this is NOT nested since the nested key is always "obs"
            cat_vectors[modality] = obs_fields


@doc_args(SLOTS_DETAILS_DOCSTRING)
class MuDataCurator(SlotsCurator):
    """Curator for `MuData`.

    {}

    Args:
        dataset: The MuData-like object to validate & annotate.
        schema: A :class:`~lamindb.Schema` object that defines the validation constraints.

    Example:
        .. literalinclude:: scripts/curate_mudata.py
            :language: python
            :caption: curate_mudata.py

    See Also:
        :meth:`~lamindb.Artifact.from_mudata`.
    """

    def __init__(
        self,
        dataset: MuData | Artifact,
        schema: Schema,
    ) -> None:
        super().__init__(dataset=dataset, schema=schema)
        if not data_is_scversedatastructure(self._dataset, "MuData"):
            raise InvalidArgument("dataset must be MuData-like.")
        if schema.otype != "MuData":
            raise InvalidArgument("Schema otype must be 'MuData'.")

        for slot, slot_schema in schema.slots.items():
            # Handle slots: "mdata.uns", "modality:uns"
            if "uns" in slot:
                df, modality, modality_slot = _handle_dict_slots(self._dataset, slot)
            else:
                # Handle slots: "modality:obs", "modality:var"
                parts = slot.split(":")
                if len(parts) == 2:
                    modality, modality_slot = parts
                    try:
                        schema_dataset = self._dataset[modality]
                        df = getattr(schema_dataset, modality_slot.rstrip(".T"))
                    except KeyError:
                        raise InvalidArgument(
                            f"Modality '{modality}' not found in MuData"
                        ) from None
                    except AttributeError:
                        raise InvalidArgument(
                            f"Attribute '{modality_slot}' not found on modality '{modality}'"
                        ) from None
                else:
                    # Handle slots: "mdata:obs", "mdata:var" (uns is a dictionary and gets handled above)
                    modality, modality_slot = None, slot
                    schema_dataset = self._dataset
                    df = getattr(schema_dataset, modality_slot.rstrip(".T"))

            # Transpose var if necessary
            if modality_slot == "var" and schema.slots[slot].itype not in {
                None,
                "Feature",
            }:
                logger.warning(
                    "auto-transposed `var` for backward compat, please indicate transposition in the schema definition by calling out `.T`: slots={'var.T': itype=bt.Gene.ensembl_gene_id}"
                )
                df = df.T
            elif modality_slot == "var.T":
                df = df.T

            self._slots[slot] = ComponentCurator(df, slot_schema, slot=slot)

            _assign_var_fields_categoricals_multimodal(
                modality=modality,
                slot_type=modality_slot,
                slot=slot,
                slot_schema=slot_schema,
                var_fields=self._var_fields,
                cat_vectors=self._cat_vectors,
                slots=self._slots,
            )

        self._columns_field = self._var_fields


@doc_args(SLOTS_DETAILS_DOCSTRING)
class SpatialDataCurator(SlotsCurator):
    """Curator for `SpatialData`.

    {}

    Args:
        dataset: The SpatialData-like object to validate & annotate.
        schema: A :class:`~lamindb.Schema` object that defines the validation constraints.

    Example:
        .. literalinclude:: scripts/curate_spatialdata.py
            :language: python
            :caption: curate_spatialdata.py

    See Also:
        :meth:`~lamindb.Artifact.from_spatialdata`.
    """

    def __init__(
        self,
        dataset: SpatialData | Artifact,
        schema: Schema,
    ) -> None:
        super().__init__(dataset=dataset, schema=schema)
        if not data_is_scversedatastructure(self._dataset, "SpatialData"):
            raise InvalidArgument("dataset must be SpatialData-like.")
        if schema.otype != "SpatialData":
            raise InvalidArgument("Schema otype must be 'SpatialData'.")

        for slot, slot_schema in schema.slots.items():
            # Handle slots: "sdata:attrs"
            if slot.startswith("attrs"):
                df, table_key, table_slot = _handle_dict_slots(self._dataset, slot)
            else:
                parts = slot.split(":")
                # Handle slots: "tables:table_key:obs", "tables:table_key:var"
                if len(parts) == 3 and parts[0] == "tables":
                    table_key, table_slot = parts[1], parts[2]
                    try:
                        slot_object = self._dataset.tables[table_key]
                        df = getattr(slot_object, table_slot.rstrip(".T"))
                    except KeyError:
                        raise InvalidArgument(
                            f"Table '{table_key}' not found in sdata.tables"
                        ) from None
                    except AttributeError:
                        raise InvalidArgument(
                            f"Attribute '{table_slot}' not found on table '{table_key}'"
                        ) from None
                else:
                    # Handle legacy single keys for backward compatibility
                    if len(parts) == 1 and parts[0] != "attrs":
                        logger.warning(
                            f"please prefix slot {slot} with 'attrs:' going forward"
                        )
                        try:
                            df = pd.DataFrame([self._dataset.attrs[slot]])
                            table_key = None
                            table_slot = slot
                        except KeyError:
                            raise InvalidArgument(
                                f"Slot '{slot}' not found in sdata.attrs"
                            ) from None
                    else:
                        raise InvalidArgument(f"Unrecognized slot format: {slot}")

            # Handle var transposition logic
            if table_slot == "var" and schema.slots[slot].itype not in {
                None,
                "Feature",
            }:
                logger.warning(
                    "auto-transposed `var` for backward compat, please indicate transposition in the schema definition by calling out `.T`: slots={'var.T': itype=bt.Gene.ensembl_gene_id}"
                )
                df = df.T
            elif table_slot == "var.T":
                df = df.T

            self._slots[slot] = ComponentCurator(df, slot_schema, slot)

            _assign_var_fields_categoricals_multimodal(
                modality=table_key,
                slot_type=table_slot,
                slot=slot,
                slot_schema=slot_schema,
                var_fields=self._var_fields,
                cat_vectors=self._cat_vectors,
                slots=self._slots,
            )

        self._columns_field = self._var_fields


@doc_args(SLOTS_DETAILS_DOCSTRING)
class TiledbsomaExperimentCurator(SlotsCurator):
    """Curator for `tiledbsoma.Experiment`.

    {}

    Args:
        dataset: The `tiledbsoma.Experiment` object.
        schema: A :class:`~lamindb.Schema` object that defines the validation constraints.

    Example:

        .. literalinclude:: scripts/curate_soma_experiment.py
            :language: python
            :caption: curate_soma_experiment.py

    See Also:
        :meth:`~lamindb.Artifact.from_tiledbsoma`.
    """

    def __init__(
        self,
        dataset: SOMAExperiment | Artifact,
        schema: Schema,
    ) -> None:
        super().__init__(dataset=dataset, schema=schema)
        if not data_is_soma_experiment(self._dataset):
            raise InvalidArgument("dataset must be SOMAExperiment-like.")
        if schema.otype != "tiledbsoma":
            raise InvalidArgument("Schema otype must be 'tiledbsoma'.")

        for slot, slot_schema in schema.slots.items():
            if slot.startswith("ms:"):
                _, modality_slot = slot.split(":")
                schema_dataset = (
                    self._dataset.ms[modality_slot.removesuffix(".T")]
                    .var.read()
                    .concat()
                    .to_pandas()
                    .drop("soma_joinid", axis=1, errors="ignore")
                )

                self._slots[slot] = ComponentCurator(
                    (schema_dataset.T if modality_slot == "var.T" else schema_dataset),
                    slot_schema,
                )
            else:
                # global Experiment obs slot
                modality_slot = slot
                schema_dataset = (
                    self._dataset.obs.read()
                    .concat()
                    .to_pandas()
                    .drop(["soma_joinid", "obs_id"], axis=1, errors="ignore")
                )
                self._slots[slot] = ComponentCurator(
                    schema_dataset,
                    slot_schema,
                )

            _assign_var_fields_categoricals_multimodal(
                modality=slot,  # not passing `measurement` here because it's a constant. The slot has the actual modality
                slot_type=modality_slot,
                slot=slot,
                slot_schema=slot_schema,
                var_fields=self._var_fields,
                cat_vectors=self._cat_vectors,
                slots=self._slots,
            )
        self._columns_field = self._var_fields


class CatVector:
    """Vector with categorical values."""

    def __init__(
        self,
        values_getter: Callable
        | Iterable[str],  # A callable or iterable that returns the values to validate.
        field: FieldAttr,  # The field to validate against.
        key: str,  # The name of the vector to validate. Only used for logging.
        values_setter: Callable | None = None,  # A callable that sets the values.
        source: SQLRecord | None = None,  # The ontology source to validate against.
        feature: Feature | None = None,
        cat_manager: DataFrameCatManager | None = None,
        filter_str: str = "",
        subtypes_list: list[str] = None,
        maximal_set: bool = True,  # whether unvalidated categoricals cause validation failure.
    ) -> None:
        if subtypes_list is None:
            subtypes_list = []
        self._values_getter = values_getter
        self._values_setter = values_setter
        self._field = field
        self._key = key
        self._source = source
        self._organism = None
        self._validated: None | list[str] = None
        self._non_validated: None | list[str] = None
        self._synonyms: None | dict[str, str] = None
        self._filter_str = filter_str
        self._subtypes_list = subtypes_list
        self._subtype_query_set = None
        self._cat_manager = cat_manager
        self.feature = feature
        self.records = None
        self._maximal_set = maximal_set
        self._type_record = None

        self._all_filters = {"source": self._source, "organism": self._organism}

        if self._filter_str:
            self._all_filters.update(
                resolve_relation_filters(
                    parse_filter_string(self._filter_str), self._field.field.model
                )  # type: ignore
            )

        # get the dtype associated record based on the nested subtypes
        if self._subtypes_list:
            type_filters = {"name": self._subtypes_list[-1]}
            if len(self._subtypes_list) > 1:
                for i, nested_subtype in enumerate(reversed(self._subtypes_list[:-1])):
                    filter_key = f"{'type__' * (i + 1)}name"
                    type_filters[filter_key] = nested_subtype
            try:
                self._type_record = self._field.field.model.get(**type_filters)
            except Exception as e:
                raise InvalidArgument(
                    f"Error retrieving type record with filters {type_filters} for field {self._field.field.name}."
                ) from e

        if hasattr(field.field.model, "_name_field"):
            label_ref_is_name = field.field.name == field.field.model._name_field
        else:
            label_ref_is_name = field.field.name == "name"
        self.label_ref_is_name = label_ref_is_name

    @property
    def values(self):
        """Get the current values using the getter function."""
        if callable(self._values_getter):
            return self._values_getter()
        return self._values_getter

    @values.setter
    def values(self, new_values):
        """Set new values using the setter function if available."""
        if callable(self._values_setter):
            self._values_setter(new_values)
        else:
            # If values_getter is not callable, it's a direct reference we can update
            self._values_getter = new_values

    @property
    def is_validated(self) -> bool:
        """Whether the vector is validated."""
        # if nothing was validated, something likely is fundamentally wrong
        # should probably add a setting `at_least_one_validated`
        result = True
        if len(self.values) > 0 and len(self.values) == len(self._non_validated):
            logger.warning(f"no values were validated for {self._key}!")
        # len(self._non_validated) != 0
        #     if maximal_set is True, return False
        #     if maximal_set is False, return True
        # len(self._non_validated) == 0
        #     return True
        if len(self._non_validated) != 0:
            if self._maximal_set:
                result = False
        return result

    def _replace_synonyms(self) -> list[str]:
        """Replace synonyms in the vector with standardized values."""

        def process_value(value, syn_mapper):
            """Helper function to process values recursively."""
            if isinstance(value, list):
                # Handle list - recursively process each item
                return [process_value(item, syn_mapper) for item in value]
            else:
                # Handle single value
                return syn_mapper.get(value, value)

        syn_mapper = self._synonyms
        # replace the values in df
        std_values = self.values.map(
            lambda unstd_val: process_value(unstd_val, syn_mapper)
        )
        # remove the standardized values from self.non_validated
        non_validated = [i for i in self._non_validated if i not in syn_mapper]
        if len(non_validated) == 0:
            self._non_validated = []
        else:
            self._non_validated = non_validated  # type: ignore
        # logging
        n = len(syn_mapper)
        if n > 0:
            syn_mapper_print = _format_values(
                [f'"{k}" → "{v}"' for k, v in syn_mapper.items()], sep=""
            )
            s = "s" if n > 1 else ""
            logger.success(
                f'standardized {n} synonym{s} in "{self._key}": {colors.green(syn_mapper_print)}'
            )
        return std_values

    def __repr__(self) -> str:
        if self._non_validated is None:
            status = "unvalidated"
        else:
            status = (
                "validated"
                if len(self._non_validated) == 0
                else f"non-validated ({len(self._non_validated)})"
            )

        field_name = getattr(self._field, "name", str(self._field))
        values_count = len(self.values) if hasattr(self.values, "__len__") else "?"
        return f"CatVector(key='{self._key}', field='{field_name}', values={values_count}, {status})"

    def _add_validated(self) -> tuple[list, list]:
        """Save features or labels records in the default instance."""
        from lamindb.models.save import save as ln_save

        registry = self._field.field.model
        field_name = self._field.field.name
        model_field = registry.__get_name_with_module__()
        filter_kwargs = get_current_filter_kwargs(registry, self._all_filters)

        valid_from_values_kwargs = {}
        for key, value in filter_kwargs.items():
            if key in {"field", "organism", "source", "mute"}:
                valid_from_values_kwargs[key] = value
            elif hasattr(registry, key) and "__" not in key:
                valid_from_values_kwargs[key] = value

        values = [
            i
            for i in self.values
            if (isinstance(i, str) and i)
            or (isinstance(i, list) and i)
            or (isinstance(i, np.ndarray) and i.size > 0)
        ]
        if not values:
            return [], []

        # if a value is a list, we need to flatten it
        str_values = _flatten_unique(values)

        # inspect the default instance and save validated records from public
        if self._type_record is not None:
            related_name = registry._meta.get_field("type").remote_field.related_name
            if registry.__name__ == "Record":
                self._subtype_query_set = self._type_record.query_records()
            else:
                self._subtype_query_set = getattr(self._type_record, related_name).all()
            values_array = np.array(str_values)
            validated_mask = self._subtype_query_set.validate(  # type: ignore
                values_array, field=self._field, **filter_kwargs, mute=True
            )
            validated_labels, non_validated_labels = (
                values_array[validated_mask],
                values_array[~validated_mask],
            )
            records = registry.from_values(
                validated_labels,
                field=self._field,
                **valid_from_values_kwargs,
                mute=True,
            )
        else:
            existing_and_public_records = registry.from_values(
                str_values, field=self._field, **valid_from_values_kwargs, mute=True
            )
            existing_and_public_labels = [
                getattr(r, field_name) for r in existing_and_public_records
            ]
            # public records that are not already in the database
            public_records = [r for r in existing_and_public_records if r._state.adding]
            # here we check to only save the public records if they are from the specified source
            # we check the uid because r.source and source can be from different instances
            if self._source:
                public_records = [
                    r for r in public_records if r.source.uid == self._source.uid
                ]
            if len(public_records) > 0:
                logger.info(f"saving validated records of '{self._key}'")
                ln_save(public_records)
                labels_saved_public = [getattr(r, field_name) for r in public_records]
                # log the saved public labels
                # the term "transferred" stresses that this is always in the context of transferring
                # labels from a public ontology or a different instance to the present instance
                if len(labels_saved_public) > 0:
                    s = "s" if len(labels_saved_public) > 1 else ""
                    logger.success(
                        f'added {len(labels_saved_public)} record{s} {colors.green("from_public")} with {model_field} for "{self._key}": {_format_values(labels_saved_public)}'
                    )
                    # non-validated records from the default instance
            non_validated_labels = [
                i for i in str_values if i not in existing_and_public_labels
            ]
            validated_labels = existing_and_public_labels
            records = existing_and_public_records

        self.records = records
        # validated, non-validated
        return validated_labels, non_validated_labels

    def _add_new(
        self,
        values: list[str],
        df: pd.DataFrame | None = None,  # remove when all users use schema
        dtype: str | None = None,
        **create_kwargs,
    ) -> None:
        """Add new labels to the registry."""
        from lamindb.models.save import save as ln_save

        registry = self._field.field.model
        field_name = self._field.field.name
        non_validated_records: SQLRecordList[Any] = []  # type: ignore
        if df is not None and registry == Feature:
            nonval_columns = Feature.inspect(df.columns, mute=True).non_validated
            non_validated_records = Feature.from_dataframe(df.loc[:, nonval_columns])
        else:
            if (
                self._organism
                and hasattr(registry, "organism")
                and registry._meta.get_field("organism").is_relation
            ):
                # make sure organism record is saved to the current instance
                create_kwargs["organism"] = _save_organism(name=self._organism)

            for value in values:
                init_kwargs = {field_name: value}
                if registry == Feature:
                    init_kwargs["dtype"] = "cat" if dtype is None else dtype
                if self._type_record is not None:
                    # if type_record is set, we need to set the type for new records
                    init_kwargs["type"] = self._type_record
                # here we create non-validated records skipping validation since we already ensured that they don't exist
                non_validated_records.append(
                    registry(**init_kwargs, **create_kwargs, _skip_validation=True)
                )
        if len(non_validated_records) > 0:
            ln_save(non_validated_records)
            model_field = colors.italic(registry.__get_name_with_module__())
            s = "s" if len(values) > 1 else ""
            logger.success(
                f'added {len(values)} record{s} with {model_field} for "{self._key}": {_format_values(values)}'
            )

    def _validate(
        self,
        values: list[str],
    ) -> tuple[list[str], dict]:
        """Validate ontology terms using LaminDB registries."""
        registry = self._field.field.model
        field_name = self._field.field.name
        model_field = f"{registry.__name__}.{field_name}"

        kwargs_current = get_current_filter_kwargs(registry, self._all_filters)

        valid_inspect_kwargs = {}
        for key, value in kwargs_current.items():
            if key in {"field", "organism", "source", "mute", "from_source"}:
                valid_inspect_kwargs[key] = value
            elif hasattr(registry, key) and "__" not in key:
                valid_inspect_kwargs[key] = value

        # inspect values from the default instance, excluding public
        registry_or_queryset = registry
        if self._subtype_query_set is not None:
            registry_or_queryset = self._subtype_query_set
        inspect_result = registry_or_queryset.inspect(
            values,
            field=self._field,
            mute=True,
            from_source=False,
            **valid_inspect_kwargs,
        )
        non_validated = inspect_result.non_validated
        syn_mapper = inspect_result.synonyms_mapper

        # inspect the non-validated values from public (BioRecord only)
        values_validated = []
        if hasattr(registry, "public"):
            public_records = registry.from_values(
                non_validated,
                field=self._field,
                mute=True,
                **valid_inspect_kwargs,
            )
            values_validated += [getattr(r, field_name) for r in public_records]

        # logging messages
        if self._cat_manager is not None:
            slot = self._cat_manager._slot
        else:
            slot = None
        in_slot = f" in slot '{slot}'" if slot is not None else ""
        slot_prefix = f".slots['{slot}']" if slot is not None else ""
        non_validated_hint_print = (
            f"curator{slot_prefix}.cat.add_new_from('{self._key}')"
        )
        non_validated = [i for i in non_validated if i not in values_validated]
        n_non_validated = len(non_validated)
        if n_non_validated == 0:
            logger.success(
                f'"{self._key}" is validated against {colors.italic(model_field)}'
            )
            return [], {}
        else:
            s = "" if n_non_validated == 1 else "s"
            print_values = _format_values(non_validated)
            warning_message = f"{colors.red(f'{n_non_validated} term{s}')} not validated in feature '{self._key}'{in_slot}: {colors.red(print_values)}\n"
            if syn_mapper:
                s = "" if len(syn_mapper) == 1 else "s"
                syn_mapper_print = _format_values(
                    [f'"{k}" → "{v}"' for k, v in syn_mapper.items()], sep=""
                )
                hint_msg = f'.standardize("{self._key}")'
                warning_message += f"    {colors.yellow(f'{len(syn_mapper)} synonym{s}')} found: {colors.yellow(syn_mapper_print)}\n    → curate synonyms via: {colors.cyan(hint_msg)}"
            if n_non_validated > len(syn_mapper):
                if syn_mapper:
                    warning_message += "\n    for remaining terms:\n"
                check_organism = ""
                if registry.__base__.__name__ == "BioRecord":
                    import bionty as bt
                    from bionty._organism import is_organism_required

                    if is_organism_required(registry):
                        organism = (
                            valid_inspect_kwargs.get("organism", False)
                            or bt.settings.organism.name
                        )
                        check_organism = f"fix organism '{organism}', "
                warning_message += f"    → {check_organism}fix typos, remove non-existent values, or save terms via: {colors.cyan(non_validated_hint_print)}"
                if self._subtype_query_set is not None:
                    warning_message += f"\n    → a valid label for subtype '{self._subtypes_list[-1]}' has to be one of {self._subtype_query_set.to_list('name')}"
            logger.info(f'mapping "{self._key}" on {colors.italic(model_field)}')
            logger.warning(warning_message)
            if self._cat_manager is not None:
                self._cat_manager._validate_category_error_messages = strip_ansi_codes(
                    warning_message
                )
            return non_validated, syn_mapper

    def validate(self) -> None:
        """Validate the vector."""
        # add source-validated values to the registry
        self._validated, self._non_validated = self._add_validated()
        self._non_validated, self._synonyms = self._validate(values=self._non_validated)

    def standardize(self) -> None:
        """Standardize the vector."""
        registry = self._field.field.model
        if not hasattr(registry, "standardize"):
            return self.values
        if self._synonyms is None:
            self.validate()
        # get standardized values
        std_values = self._replace_synonyms()
        # update non_validated values
        self._non_validated = [
            i for i in self._non_validated if i not in self._synonyms.keys()
        ]
        # remove synonyms since they are now standardized
        self._synonyms = {}
        # update the values with the standardized values
        self.values = std_values

    def add_new(self, **create_kwargs) -> None:
        """Add new values to the registry."""
        if self._non_validated is None:
            self.validate()
        if len(self._synonyms) > 0:
            # raise error because .standardize modifies the input dataset
            raise ValidationError(
                "Please run `.standardize()` before adding new values."
            )
        self._add_new(
            values=self._non_validated,
            **create_kwargs,
        )
        # remove the non_validated values since they are now registered
        self._non_validated = []


class DataFrameCatManager:
    """Manage categoricals by updating registries.

    This class is accessible from within a `DataFrameCurator` via the `.cat` attribute.

    If you find non-validated values, you have two options:

    - new values found in the data can be registered via `DataFrameCurator.cat.add_new_from()` :meth:`~lamindb.curators.core.DataFrameCatManager.add_new_from`
    - non-validated values can be accessed via `DataFrameCurator.cat.add_new_from()` :meth:`~lamindb.curators.core.DataFrameCatManager.non_validated` and addressed manually
    """

    def __init__(
        self,
        df: pd.DataFrame | Artifact,
        columns_field: FieldAttr = Feature.name,
        categoricals: list[Feature] | None = None,
        sources: dict[str, SQLRecord] | None = None,
        index: Feature | None = None,
        slot: str | None = None,
        maximal_set: bool = False,
    ) -> None:
        self._non_validated = None
        self._index = index
        self._artifact: Artifact = None  # pass the dataset as an artifact
        self._dataset: Any = df  # pass the dataset as a UPathStr or data object
        if isinstance(self._dataset, Artifact):
            self._artifact = self._dataset
            self._dataset = self._dataset.load(is_run_input=False)
        self._is_validated: bool = False
        self._categoricals = categoricals or []
        self._non_validated = None
        self._sources = sources or {}
        self._columns_field = columns_field
        self._validate_category_error_messages: str = ""
        self._cat_vectors: dict[str, CatVector] = {}
        self._slot = slot
        self._maximal_set = maximal_set

        self._cat_vectors["columns"] = CatVector(
            values_getter=lambda: self._dataset.keys(),  # lambda ensures the inplace update
            values_setter=lambda new_values: setattr(
                self._dataset, "columns", pd.Index(new_values)
            )
            if isinstance(self._dataset, pd.DataFrame)
            else None,
            field=columns_field,
            key="columns" if isinstance(self._dataset, pd.DataFrame) else "keys",
            source=self._sources.get("columns"),
            cat_manager=self,
            maximal_set=self._maximal_set,
        )
        for feature in self._categoricals:
            result = parse_dtype(feature.dtype)[
                0
            ]  # TODO: support composite dtypes for categoricals
            key = feature.name
            self._cat_vectors[key] = CatVector(
                values_getter=lambda k=key: self._dataset[
                    k
                ],  # Capture key as default argument
                values_setter=lambda new_values, k=key: self._dataset.__setitem__(
                    k, new_values
                ),
                field=result["field"],
                key=key,
                source=self._sources.get(key),
                feature=feature,
                cat_manager=self,
                filter_str=result["filter_str"],
                subtypes_list=result["subtypes_list"],
            )
        if index is not None and index.dtype.startswith("cat"):
            result = parse_dtype(index.dtype)[0]
            key = "index"
            self._cat_vectors[key] = CatVector(
                values_getter=self._dataset.index,
                values_setter=lambda new_values: setattr(
                    self._dataset, "index", new_values
                ),
                field=result["field"],
                key=key,
                feature=index,
                cat_manager=self,
                filter_str=result["filter_str"],
                subtypes_list=result["subtypes_list"],
            )

    @property
    def non_validated(self) -> dict[str, list[str]]:
        """Return the non-validated features and labels."""
        if self._non_validated is None:
            raise ValidationError("Please run validate() first!")
        return {
            key: cat_vector._non_validated
            for key, cat_vector in self._cat_vectors.items()
            if cat_vector._non_validated and key != "columns"
        }

    @property
    def categoricals(self) -> list[Feature]:
        """The categorical features."""
        return self._categoricals

    def __repr__(self) -> str:
        cls_name = colors.green(self.__class__.__name__)

        status_str = (
            f"{colors.green('validated')}"
            if self._is_validated
            else f"{colors.yellow('unvalidated')}"
        )

        info_parts = []

        cat_count = len(self._categoricals)
        if cat_count > 0:
            info_parts.append(f"categorical_features={cat_count}")

        if self._slot:
            info_parts.append(f"slot: {colors.italic(self._slot)}")

        info_str = ", ".join(info_parts)
        if info_str:
            return f"{cls_name}({info_str}, {status_str})"
        else:
            return f"{cls_name}({status_str})"

    def lookup(self, public: bool = False) -> CatLookup:
        """Lookup categories.

        Args:
            public: If "public", the lookup is performed on the public reference.
        """
        return CatLookup(
            categoricals=self._categoricals,
            slots={"columns": self._columns_field},
            public=public,
            sources=self._sources,
        )

    def validate(self) -> bool:
        """Validate variables and categorical observations."""
        self._validate_category_error_messages = ""  # reset the error messages

        validated = True
        for key, cat_vector in self._cat_vectors.items():
            logger.info(f"validating vector {key}")
            cat_vector.validate()
            validated &= cat_vector.is_validated
        self._is_validated = validated
        self._non_validated = {}  # type: ignore

        if self._index is not None:
            # cat_vector.validate() populates validated labels
            # the index should become part of the feature set corresponding to the dataframe
            if self._cat_vectors["columns"].records is not None:
                self._cat_vectors["columns"].records.insert(0, self._index)  # type: ignore
            else:
                self._cat_vectors["columns"].records = [self._index]  # type: ignore

        return self._is_validated

    def standardize(self, key: str) -> None:
        """Replace synonyms with standardized values.

        Modifies the input dataset inplace.

        Args:
            key: The key referencing the column in the DataFrame to standardize.
        """
        if self._artifact is not None:
            raise RuntimeError(
                "Cannot mutate the dataset when an artifact is passed! Please load the dataset into memory using `dataset.load()` and pass it to a curator."
            )

        if key == "all":
            logger.warning(
                "'all' is deprecated, please pass a single key from `.non_validated.keys()` instead!"
            )
            for k in self.non_validated.keys():
                self._cat_vectors[k].standardize()
        else:
            self._cat_vectors[key].standardize()

    def add_new_from(self, key: str, **kwargs):
        """Add validated & new categories.

        Args:
            key: The key referencing the slot in the DataFrame from which to draw terms.
            **kwargs: Additional keyword arguments to pass to create new records
        """
        if len(kwargs) > 0 and key == "all":
            raise ValueError("Cannot pass additional arguments to 'all' key!")
        if key == "all":
            logger.warning(
                "'all' is deprecated, please pass a single key from `.non_validated.keys()` instead!"
            )
            for k in self.non_validated.keys():
                self._cat_vectors[k].add_new(**kwargs)
        else:
            self._cat_vectors[key].add_new(**kwargs)


def get_current_filter_kwargs(
    registry: type[SQLRecord], kwargs: dict[str, SQLRecord]
) -> dict:
    """Make sure the source and organism are saved in the same database as the registry."""
    db = registry.filter().db
    filter_kwargs = kwargs.copy()

    for key, value in kwargs.items():
        if isinstance(value, SQLRecord) and value._state.db != "default":
            if db is None or db == "default":
                value_default = copy.copy(value)
                value_default.save()
                filter_kwargs[key] = value_default

    return filter_kwargs


def get_organism_kwargs(
    field: FieldAttr, organism: str | None = None, values: Any = None
) -> dict[str, str]:
    """Check if a registry needs an organism and return the organism name."""
    registry = field.field.model
    if registry.__base__.__name__ == "BioRecord":
        import bionty as bt
        from bionty._organism import is_organism_required

        from ..models._from_values import get_organism_record_from_field

        if is_organism_required(registry):
            if organism is not None or bt.settings.organism is not None:
                return {"organism": organism or bt.settings.organism.name}
            else:
                organism_record = get_organism_record_from_field(
                    field, organism=organism, values=values
                )
                if organism_record is not None:
                    return {"organism": organism_record.name}
    return {}


def annotate_artifact(
    artifact: Artifact,
    *,
    curator: SlotsCurator | None = None,
    cat_vectors: dict[str, CatVector] | None = None,
) -> Artifact:
    from .. import settings
    from ..models.artifact import add_labels

    if cat_vectors is None:
        cat_vectors = {}

    # annotate with labels
    for key, cat_vector in cat_vectors.items():
        if (
            cat_vector._field.field.model == Feature
            or key == "columns"
            or key == "var_index"
            or cat_vector.records is None
        ):
            continue
        if len(cat_vector.records) > settings.annotation.n_max_records:
            logger.important(
                f"not annotating with {len(cat_vector.records)} labels for feature {key} as it exceeds {settings.annotation.n_max_records} (ln.settings.annotation.n_max_records)"
            )
            continue
        add_labels(
            artifact,
            records=cat_vector.records,
            feature=cat_vector.feature,
            feature_ref_is_name=None,  # do not need anymore
            label_ref_is_name=cat_vector.label_ref_is_name,
            from_curator=True,
        )

    # annotate with inferred schemas aka feature sets
    if (
        artifact.otype == "DataFrame" and getattr(curator, "_schema", None) is None
    ):  # Prevent overwriting user-defined schemas that contain slots
        features = cat_vectors["columns"].records
        if features is not None:
            index_feature = artifact.schema.index
            feature_set = Schema(
                features=[f for f in features if f != index_feature],
                itype=artifact.schema.itype,
                index=index_feature,
                minimal_set=artifact.schema.minimal_set,
                maximal_set=artifact.schema.maximal_set,
                coerce_dtype=artifact.schema.coerce_dtype,
                ordered_set=artifact.schema.ordered_set,
            )
            if (
                feature_set._state.adding
                and len(features) > settings.annotation.n_max_records
            ):
                logger.important(
                    f"not annotating with {len(features)} features as it exceeds {settings.annotation.n_max_records} (ln.settings.annotation.n_max_records)"
                )
                itype = (
                    Feature.name
                    if artifact.schema.itype == "Composite"
                    else parse_cat_dtype(artifact.schema.itype, is_itype=True)["field"]
                )
                feature_set = Schema(itype=itype, n=len(features))
            artifact.feature_sets.add(
                feature_set.save(), through_defaults={"slot": "columns"}
            )
    else:
        for slot, slot_curator in curator._slots.items():
            # var_index is backward compat (2025-05-01)
            name = (
                "var_index"
                if (slot == "var" and "var_index" in slot_curator.cat._cat_vectors)
                else "columns"
            )
            features = slot_curator.cat._cat_vectors[name].records
            if features is None:
                logger.warning(f"no features found for slot {slot}")
                continue
            validating_schema = slot_curator._schema
            index_feature = validating_schema.index
            feature_set = Schema(
                features=[f for f in features if f != index_feature],
                itype=validating_schema.itype,
                index=index_feature,
                minimal_set=validating_schema.minimal_set,
                maximal_set=validating_schema.maximal_set,
                coerce_dtype=validating_schema.coerce_dtype,
                ordered_set=validating_schema.ordered_set,
            )
            if (
                feature_set._state.adding
                and len(features) > settings.annotation.n_max_records
            ):
                logger.important(
                    f"not annotating with {len(features)} features for slot {slot} as it exceeds {settings.annotation.n_max_records} (ln.settings.annotation.n_max_records)"
                )
                itype = (
                    Feature.name
                    if artifact.schema.slots[slot].itype == "Composite"
                    else parse_cat_dtype(
                        artifact.schema.slots[slot].itype, is_itype=True
                    )["field"]
                )
                feature_set = Schema(itype=itype, n=len(features))
            artifact.feature_sets.add(
                feature_set.save(), through_defaults={"slot": slot}
            )

    slug = ln_setup.settings.instance.slug
    if ln_setup.settings.instance.is_remote:  # pdagma: no cover
        ui_url = ln_setup.settings.instance.ui_url
        logger.important(f"go to {ui_url}/{slug}/artifact/{artifact.uid}")
    return artifact


def _flatten_unique(series: pd.Series[list[Any] | Any]) -> list[Any]:
    """Flatten a Pandas series containing lists or single items into a unique list of elements.

    The order of elements in the result list preserves the order they first appear in the input series.
    """
    # Use dict.fromkeys to preserve order while ensuring uniqueness
    result: dict = {}

    for item in series:
        if isinstance(item, list | np.ndarray):
            # Add each element to the dict (only first occurrence is kept)
            for element in item:
                result[element] = None
        else:
            result[item] = None

    # Return the keys as a list, preserving order
    return list(result.keys())


def _save_organism(name: str):
    """Save an organism record."""
    import bionty as bt

    organism = bt.Organism.filter(name=name).one_or_none()
    if organism is None:
        organism = bt.Organism.from_source(name=name)
        if organism is None:
            raise ValidationError(
                f'Organism "{name}" not found from public reference\n'
                f'      → please save it from a different source: bt.Organism.from_source(name="{name}", source).save()'
                f'      → or manually save it without source: bt.Organism(name="{name}").save()'
            )
        organism.save()
    return organism



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

import zarr
from anndata import __version__ as anndata_version
from lamin_utils import logger
from lamindb_setup.core.upath import LocalPathClasses, S3FSMap, UPath, create_mapper
from packaging import version

from lamindb.core._compat import with_package

if version.parse(anndata_version) < version.parse("0.11.0"):
    from anndata._io import read_zarr as read_anndata_zarr
else:
    from anndata.io import read_zarr as read_anndata_zarr

if version.parse(zarr.__version__) >= version.parse("3.0.0a0"):
    IS_ZARR_V3 = True
    from zarr.abc.store import Store
else:
    IS_ZARR_V3 = False
    from zarr.storage import Store  # noqa

if TYPE_CHECKING:
    from fsspec import FSMap
    from lamindb_setup.types import UPathStr

    from lamindb.core.types import ScverseDataStructures


def get_zarr_store(
    path: UPathStr, *, check: bool = False, create: bool = False
) -> str | S3FSMap | FSMap | Store:
    """Creates the correct object that can be used to open a zarr file depending on local or remote location."""
    storepath, storepath_str = UPath(path), str(path)
    if isinstance(storepath, LocalPathClasses):
        store = storepath_str
    elif IS_ZARR_V3:
        # todo: also check how to treat non-asynchronous filesystems
        # zarr has something for this, using fsspec async wrapper
        # check FsspecStore code
        store = zarr.storage.FsspecStore.from_upath(UPath(storepath, asynchronous=True))
    else:
        store = create_mapper(storepath.fs, storepath_str, check=check, create=create)

    return store


def _identify_zarr_type_from_storage(
    storage: zarr.Group,
) -> Literal["anndata", "mudata", "spatialdata", "unknown"]:
    """Internal helper to identify zarr type from an open storage object."""
    try:
        if storage.attrs.get("encoding-type", "") == "anndata":
            return "anndata"
        elif storage.attrs.get("encoding-type", "") == "MuData":
            return "mudata"
        elif "spatialdata_attrs" in storage.attrs:
            return "spatialdata"
    except Exception as error:
        logger.warning(f"an exception occurred {error}")
    return "unknown"


def identify_zarr_type(
    storepath: UPathStr, *, check: bool = True
) -> Literal["anndata", "mudata", "spatialdata", "unknown"]:
    """Identify whether a zarr store is AnnData, SpatialData, or unknown type."""
    suffixes = UPath(storepath).suffixes
    if ".anndata" in suffixes:
        return "anndata"
    elif ".mudata" in suffixes:
        return "mudata"
    elif ".spatialdata" in suffixes:
        return "spatialdata"

    store = get_zarr_store(storepath, check=check)
    try:
        storage = zarr.open(store, mode="r")
        return _identify_zarr_type_from_storage(storage)
    except Exception as error:
        logger.warning(
            f"an exception occured while trying to open the zarr store\n {error}"
        )
    return "unknown"


def load_zarr(
    storepath: UPathStr,
    expected_type: Literal["anndata", "mudata", "spatialdata"] = None,
) -> ScverseDataStructures:
    """Loads a zarr store and returns the corresponding scverse data structure.

    Args:
        storepath: Path to the zarr store
        expected_type: If provided, ensures the zarr store is of this type ("anndata", "mudata", "spatialdata")
                       and raises ValueError if it's not
    """
    store = get_zarr_store(storepath, check=True)
    # Open the storage once
    try:
        storage = zarr.open(store, mode="r")
    except Exception as error:
        raise ValueError(f"Could not open zarr store: {error}") from None

    actual_type = _identify_zarr_type_from_storage(storage)
    if expected_type is not None and actual_type != expected_type:
        raise ValueError(
            f"Expected zarr store of type '{expected_type}', but found '{actual_type}'"
        )

    match actual_type:
        case "anndata":
            scverse_obj = read_anndata_zarr(store)
        case "mudata":
            scverse_obj = with_package("mudata", lambda mod: mod.read_zarr(store))
        case "spatialdata":
            scverse_obj = with_package("spatialdata", lambda mod: mod.read_zarr(store))
        case "unknown" | _:
            raise ValueError(
                "Unable to determine zarr store format and therefore cannot load Artifact."
            )
    return scverse_obj



import re

import lamindb as ln
import pytest
from lamindb.errors import FieldValidationError


def test_record():
    with pytest.raises(
        FieldValidationError,
        match=re.escape(
            "Only name, type, is_type, description, schema, reference, reference_type are valid keyword arguments"
        ),
    ):
        ln.Record(x=1)

    with pytest.raises(ValueError) as error:
        ln.Record(1)
    assert error.exconly() == "ValueError: Only one non-keyword arg allowed"

    with pytest.raises(
        ValueError,
        match=re.escape(
            "'my_type' should start with a capital letter given you're defining a type"
        ),
    ):
        ln.Record(name="my_type", is_type=True)


def test_record_plural_type_warning(ccaplog):
    ln.Record(name="MyThings", is_type=True)
    assert (
        "name 'MyThings' for type ends with 's', in case you're naming with plural, consider the singular for a type name"
        in ccaplog.text
    )


def test_name_lookup():
    my_type = ln.Record(name="MyType", is_type=True).save()
    label1 = ln.Record(name="label 1", type=my_type).save()
    label2 = ln.Record(name="label 1", type=my_type)
    assert label2 == label1
    label2 = ln.Record(name="label 1")
    assert label2 != label1
    label2.save()
    label3 = ln.Record(name="label 1")
    assert label3 == label2
    label2.delete(permanent=True)
    label1.delete(permanent=True)
    my_type.delete(permanent=True)



"""Types.

Central object types.

.. autosummary::
   :toctree: .

   ArtifactKind
   TransformType
   Dtype

Basic types.

.. autosummary::
   :toctree: .

   UPathStr
   StrField
   ListLike
   FieldAttr
"""

from __future__ import annotations

from typing import Literal, Union

import numpy as np
import pandas as pd
from django.db.models.query_utils import DeferredAttribute as FieldAttr
from lamindb_setup.types import UPathStr  # noqa: F401

# need to use Union because __future__.annotations doesn't do the job here <3.10
# typing.TypeAlias, >3.10 on but already deprecated
ListLike = Union[list[str], pd.Series, np.array]
StrField = Union[str, FieldAttr]  # typing.TypeAlias

TransformType = Literal["pipeline", "notebook", "script", "function", "linker"]
ArtifactKind = Literal["dataset", "model", "__lamindb_run__"]

# below is used for Feature.dtype and Param.dtype
Dtype = Literal[
    "cat",  # categoricals
    "num",  # numericals
    "str",  # string
    "int",  # integer / numpy.integer
    "float",  # float
    "bool",  # boolean
    "date",  # date
    "datetime",  # datetime
    "dict",  # dictionary
    "object",  # this is a pandas input dtype, we're only using it for complicated types, not for strings
    "path",  # path, validated as str, but specially treated in the UI
]
"""Data type.

String-serialized representations of common data types.

Overview
========

============  ============  =================================================
description   lamindb       pandas
============  ============  =================================================
categorical   `"cat"`       `category`
numerical     `"num"`       `int | float`
integer       `"int"`       `int64 | int32 | int16 | int8 | uint | ...`
float         `"float"`     `float64 | float32 | float16 | float8 | ...`
string        `"str"`       `object`
datetime      `"datetime"`  `datetime`
date          `"date"`      `object` (pandera requires an ISO-format string, convert with `df["date"] = df["date"].dt.date`)
dictionary    `"dict"`      `object`
path          `"path"`      `str` (pandas does not have a dedicated path type, validated as `str`)
============  ============  =================================================

Categoricals
============

Beyond indicating that a feature is a categorical, `lamindb` allows you to define the registry to which values are restricted.

For example, `'cat[ULabel]'` or `'cat[bionty.CellType]'` indicate that permissible values are from the `ULabel` or `CellType` registry, respectively.

You can also reference multiple registries, e.g., `'cat[ULabel|bionty.CellType]'` indicates that values can be from either registry.

You can also restrict to sub-types defined in registries via the `type` column, e.g., `'cat[ULabel[CellMedium]]'` indicates that values must be of type `CellMedium` within the `ULabel` registry.

Literal
=======

A `Dtype` object in `lamindb` is a `Literal` up to further specification of `"cat"`.

"""
FeatureDtype = Dtype  # backward compat



from __future__ import annotations

import subprocess
from pathlib import Path

from lamin_utils import logger
from lamindb_setup import settings as setup_settings
from lamindb_setup.core.hashing import hash_code

from ..core._settings import sanitize_git_repo_url, settings
from ..errors import BlobHashNotFound


def get_git_repo_from_remote(url: str | None = None, depth: int | None = 10) -> Path:
    """Clone the git repository if not already cloned.

    If `depth` is provided, a shallow clone is performed and no tags are fetched.
    """
    repo_url = url or settings.sync_git_repo
    repo_dir = setup_settings.cache_dir / repo_url.split("/")[-1]
    if repo_dir.exists():
        logger.warning(f"git repo {repo_dir} already exists locally")
        return repo_dir
    logger.important(
        f"running outside of synched git repo, cloning {repo_url} into {repo_dir}"
    )
    args = ["git", "clone", f"{repo_url}.git"]
    if depth is not None:
        # if depth is provided, will not fetch tags
        args += ["--depth", f"{depth}"]
    result = subprocess.run(
        args,
        capture_output=True,
        cwd=setup_settings.cache_dir,
    )
    if result.returncode != 0 or not repo_dir.exists():
        raise RuntimeError(result.stderr.decode())
    return repo_dir


def check_local_git_repo() -> bool:
    result = subprocess.run(
        ["git", "config", "--get remote.origin.url"],
        capture_output=True,
    )
    result_str = result.stdout.decode().strip()
    if result_str == "":
        # running-not-in-a-git-repo
        return False
    else:
        remote_url = sanitize_git_repo_url(result_str)
        if remote_url == settings.sync_git_repo:
            # running-in-correct-git-repo
            return True
        else:
            # running-outside-of-correct-git-repo
            return False


def get_git_commit_hash(blob_hash: str, repo_dir: Path | None = None) -> str | None:
    # Fetch all remote branches so that we can also search them
    fetch_command = ["git", "fetch", "origin", "+refs/heads/*:refs/remotes/origin/*"]
    subprocess.run(fetch_command, cwd=repo_dir, check=True)

    # Find the commit containing the blob hash in all branches
    command = [
        "git",
        "log",
        "--all",
        f"--find-object={blob_hash}",
        "--pretty=format:%H",
    ]
    result = subprocess.run(
        command,
        capture_output=True,
        cwd=repo_dir,
    )
    # We just care to find one commit
    # Hence, we split by new line ("\n") and use the first one
    commit_hash = result.stdout.decode().split("\n")[0]

    if not commit_hash or result.returncode == 1:
        return None

    default_branch = (
        subprocess.run(
            ["git", "rev-parse", "--abbrev-ref", "origin/HEAD"],
            capture_output=True,
            cwd=repo_dir,
            text=True,
        )
        .stdout.strip()
        .split("/")[-1]
    )

    # Find all branches containing the commit
    commit_containing_branches = subprocess.run(
        ["git", "branch", "--all", "--contains", commit_hash],
        capture_output=True,
        cwd=repo_dir,
        text=True,
    ).stdout.split("\n")

    # Clean up branch names and filter out the default branch
    commit_containing_branches = [
        branch.strip().replace("remotes/", "")
        for branch in commit_containing_branches
        if branch.strip()
    ]
    non_default_branches = [
        branch for branch in commit_containing_branches if default_branch not in branch
    ]

    if non_default_branches:
        logger.warning(
            f"code blob hash {blob_hash} was found in non-default branch(es): {', '.join(non_default_branches)}"
        )

    assert (  # noqa: S101
        len(commit_hash) == 40
    ), f"commit hash |{commit_hash}| is not 40 characters long"

    return commit_hash


def get_filepath_within_git_repo(
    commit_hash: str, blob_hash: str, repo_dir: Path | None
) -> str:
    # repo_dir might not point to the root of the
    # the git repository because git log --find-object works
    # from anywhere in the repo, hence, let's get the root
    repo_root = (
        subprocess.run(
            ["git", "rev-parse", "--show-toplevel"],
            capture_output=True,
            cwd=repo_dir,
        )
        .stdout.decode()
        .strip()
    )
    # Run the git commands separately to circumvent spawning a shell
    git_command = ["git", "ls-tree", "-r", commit_hash]
    git_process = subprocess.Popen(
        git_command,
        stdout=subprocess.PIPE,
        cwd=repo_root,
    )

    grep_command = ["grep", "-E", blob_hash]
    result = subprocess.run(
        grep_command,
        stdin=git_process.stdout,
        capture_output=True,
        cwd=repo_root,
    )

    # Close the stdout to allow git_process to receive a SIGPIPE if grep_command exits
    git_process.stdout.close()
    git_process.wait()

    command = " ".join(git_command) + " | " + " ".join(grep_command)
    if result.returncode != 0 and result.stderr.decode() != "":
        raise RuntimeError(f"{command}\n{result.stderr.decode()}")
    if len(result.stdout.decode()) == 0:
        raise RuntimeError(
            f"Could not find path in git repo {settings.sync_git_repo} running:\n{command}"
            f"\nin local clone: {repo_root}"
        )
    filepath = result.stdout.decode().split()[-1]
    return filepath


def get_transform_reference_from_git_repo(path: Path) -> str:
    blob_hash = hash_code(path).hexdigest()
    commit_hash = None
    if check_local_git_repo():
        repo_dir = None
    else:
        repo_dir = get_git_repo_from_remote()
    commit_hash = get_git_commit_hash(blob_hash, repo_dir=repo_dir)
    if commit_hash is None:
        if repo_dir is None:
            repo_dir = Path.cwd()
        raise BlobHashNotFound(
            f"❌ Did not find blob hash {blob_hash} in git repo ({settings.sync_git_repo}) {repo_dir}\n"
            f"Did you commit the script? -> {path}"
        )
    gitpath = get_filepath_within_git_repo(commit_hash, blob_hash, repo_dir)
    reference = f"{settings.sync_git_repo}/blob/{commit_hash}/{gitpath}"
    return reference


def get_and_validate_git_metadata(
    url: str,
    path: str,
    version: str | None = None,
    branch: str | None = None,
) -> tuple[str, str]:
    """Get metadata from a git repository.

    Args:
        url: Git repository URL (e.g., "https://github.com/user/repo")
        path: Path to the main script within the repository
        version: Optional version/tag to checkout
        branch: Optional branch name (defaults to repository's default branch)

    Returns:
        Dictionary containing:
            - commit_hash: The current commit hash
            - url: The repository URL
            - main_script: Path to the main script
            - revision: The version/tag (if provided)
            - branch: The branch name

    Raises:
        RuntimeError: If git operations fail
        FileNotFoundError: If the specified path does not exist in the repository
    """
    url = sanitize_git_repo_url(url)
    repo_dir = get_git_repo_from_remote(url, depth=None)

    # Determine the branch to use
    if branch is None:
        # Get the default branch if not specified
        result_str = subprocess.run(
            ["git", "rev-parse", "--abbrev-ref", "origin/HEAD"],
            capture_output=True,
            cwd=repo_dir,
            text=True,
        )
        if result_str.returncode == 0:
            branch = result_str.stdout.strip().split("/")[-1]
        else:
            branch = "main"  # fallback to main

    # Fetch the latest changes
    subprocess.run(
        ["git", "fetch", "origin"],
        capture_output=True,
        cwd=repo_dir,
        check=True,
    )

    # Checkout the specified version or branch
    if version is not None:
        # Version takes precedence - checkout the tag/version
        result = subprocess.run(
            ["git", "checkout", version],
            capture_output=True,
            cwd=repo_dir,
        )
        if result.returncode != 0:
            raise ValueError(
                f"Failed to checkout version {version}: {result.stderr.decode()}"
            )
        logger.info(f"checked out version {version}")
    else:
        # Checkout the branch
        result = subprocess.run(
            ["git", "checkout", f"origin/{branch}"],
            capture_output=True,
            cwd=repo_dir,
        )
        if result.returncode != 0:
            raise ValueError(
                f"Failed to checkout branch {branch}: {result.stderr.decode()}"
            )
        logger.info(f"checked out branch {branch}")

    # Get the current commit hash
    result_str = subprocess.run(
        ["git", "rev-parse", "HEAD"],
        capture_output=True,
        cwd=repo_dir,
        text=True,
    )
    if result_str.returncode != 0:
        raise RuntimeError(f"Failed to get commit hash: {result_str.stderr}")

    commit_hash = result_str.stdout.strip()

    assert (  # noqa: S101
        len(commit_hash) == 40
    ), f"commit hash |{commit_hash}| is not 40 characters long"

    # Verify that the path exists as a file in the repository
    file_path = repo_dir / path
    if not file_path.exists():
        raise FileNotFoundError(f"Path '{path}' does not exist in repository {url}")
    if not file_path.is_file():
        raise FileNotFoundError(
            f"Path '{path}' exists but is not a file in repository {url}"
        )
    return url, commit_hash



import datetime

import bionty as bt
import lamindb as ln
import pandas as pd
import pytest
from lamindb import Record
from lamindb.errors import ValidationError
from lamindb.models.feature import (
    parse_dtype,
    parse_filter_string,
    resolve_relation_filters,
    serialize_dtype,
)


@pytest.fixture
def source():
    source = bt.Source(
        name="test_name",
        description="test_description",
        organism="human",
        entity="bionty.Gene",
        version="2026-01-01",
    )
    source.uid = "testuid1"
    source.save()
    return source


@pytest.fixture
def organism():
    organism = bt.Organism(name="test_organism")
    organism.uid = "testuid2"
    organism.save()
    return organism


# -----------------------------------------------------------------------------
# serializing dtypes
# -----------------------------------------------------------------------------


def test_serialize_basic_dtypes():
    assert serialize_dtype(int) == "int"
    assert serialize_dtype(float) == "float"
    assert serialize_dtype(str) == "str"
    assert serialize_dtype(bool) == "bool"
    assert serialize_dtype(dict) == "dict"
    # assert serialize_dtype(bytes) == "bytes"  # not yet supported
    assert serialize_dtype(datetime.datetime) == "datetime"
    assert serialize_dtype(datetime.date) == "date"


def test_serialize_basic_list_dtypes():
    assert serialize_dtype(list[int]) == "list[int]"
    assert serialize_dtype(list[float]) == "list[float]"
    assert serialize_dtype(list[str]) == "list[str]"
    assert serialize_dtype(list[bool]) == "list[bool]"
    assert serialize_dtype(list[dict]) == "list[dict]"
    assert serialize_dtype(list[datetime.datetime]) == "list[datetime]"
    assert serialize_dtype(list[datetime.date]) == "list[date]"


def test_seralize_pandas_numpy_dtypes():
    series = pd.Series([1, 4, 0, 10, 9], dtype="uint")
    assert series.dtype.name == "uint64"
    assert serialize_dtype(series.dtype) == "int"


def test_serialize_user():
    feature = ln.Feature(
        name="user_feat", dtype="cat[User]"
    )  # calls parse_dtype() to verify
    feature = ln.Feature(name="user_feat", dtype=ln.User)  # calls serialize_dtype()
    assert feature.dtype == "cat[User]"


def test_serialize_record_objects():
    record_type_ist1 = ln.Record(name="Institute1", is_type=True).save()
    record_type_dpt1 = ln.Record(
        name="Department1", type=record_type_ist1, is_type=True
    ).save()
    record_type_lab = ln.Record(
        name="Instrument", type=record_type_dpt1, is_type=True
    ).save()
    serialized_str = "cat[Record[Institute1[Department1[Instrument]]]]"
    assert serialize_dtype(record_type_lab) == serialized_str
    record_type_lab.delete(permanent=True)
    record_type_dpt1.delete(permanent=True)
    record_type_ist1.delete(permanent=True)


def test_serialize_union_of_registries():
    serialized_str = "cat[Record|bionty.Gene]"
    assert serialize_dtype([ln.Record, bt.Gene]) == serialized_str
    serialized_str = "cat[bionty.CellType|bionty.CellLine]"
    assert serialize_dtype([bt.CellType, bt.CellLine]) == serialized_str


def test_serialize_with_field_information():
    serialized_str = "cat[bionty.Gene.ensembl_gene_id]"
    assert serialize_dtype(bt.Gene.ensembl_gene_id) == serialized_str
    serialized_str = "cat[bionty.CellType.uid|bionty.CellLine.uid]"
    assert serialize_dtype([bt.CellType.uid, bt.CellLine.uid]) == serialized_str


def test_serialize_with_additional_filters():
    pass
    # see parse_dtype
    # see parse_dtype


# -----------------------------------------------------------------------------
# parsing serialized dtypes
# -----------------------------------------------------------------------------


def test_simple_record_with_subtype_and_field():
    dtype_str = "cat[Record[Customer].name]"
    result = parse_dtype(dtype_str)
    assert len(result) == 1
    assert result[0] == {
        "registry_str": "Record",
        "filter_str": "",
        "field_str": "name",
        "registry": Record,
        "field": Record.name,
        "subtypes_list": ["Customer"],
    }


def test_multiple_records_with_subtypes_and_fields():
    dtype_str = "cat[Record[Customer].name|Record[Supplier].name]"
    result = parse_dtype(dtype_str)
    assert len(result) == 2
    assert result[0] == {
        "registry_str": "Record",
        "filter_str": "",
        "field_str": "name",
        "registry": Record,
        "field": Record.name,
        "subtypes_list": ["Customer"],
    }
    assert result[1] == {
        "registry_str": "Record",
        "filter_str": "",
        "field_str": "name",
        "registry": Record,
        "field": Record.name,
        "subtypes_list": ["Supplier"],
    }


def test_bionty_celltype_with_field():
    dtype_str = "cat[bionty.CellType.ontology_id]"
    result = parse_dtype(dtype_str)
    assert len(result) == 1
    assert result[0] == {
        "registry_str": "bionty.CellType",
        "filter_str": "",
        "field_str": "ontology_id",
        "registry": bt.CellType,
        "field": bt.CellType.ontology_id,
        "subtypes_list": [],
    }


def test_bionty_perturbations_with_field():
    dtype_str = "cat[bionty.CellType.uid|bionty.CellLine.uid]"
    result = parse_dtype(dtype_str)
    assert len(result) == 2
    assert result[0] == {
        "registry_str": "bionty.CellType",
        "filter_str": "",
        "field_str": "uid",
        "registry": bt.CellType,
        "field": bt.CellType.uid,
        "subtypes_list": [],
    }
    assert result[1] == {
        "registry_str": "bionty.CellLine",
        "filter_str": "",
        "field_str": "uid",
        "registry": bt.CellLine,
        "field": bt.CellLine.uid,
        "subtypes_list": [],
    }


def test_invalid_registry():
    dtype_str = "cat[InvalidRegistry.field]"
    with pytest.raises(ValidationError) as exc_info:
        parse_dtype(dtype_str)
    assert "invalid dtype" in str(exc_info.value)


def test_empty_category():
    dtype_str = "cat[]"
    result = parse_dtype(dtype_str)
    assert result == []


def test_malformed_categorical():
    dtype_str = "cat ? str"
    with pytest.raises(ValueError) as err:
        parse_dtype(dtype_str)
    assert err.exconly().startswith(
        f"ValueError: dtype is '{dtype_str}' but has to be one of"
    )
    dtype_str = "cat[Record[Customer.name"
    with pytest.raises(ValueError) as err:
        parse_dtype(dtype_str)
    assert err.exconly().startswith(
        f"ValueError: dtype is '{dtype_str}' but has to be one of"
    )


def test_simple_registry_without_field():
    dtype_str = "cat[Record]"
    result = parse_dtype(dtype_str)
    assert len(result) == 1
    assert result[0] == {
        "registry_str": "Record",
        "filter_str": "",
        "field_str": "name",
        "registry": Record,
        "field": Record.name,
        "subtypes_list": [],
    }


def test_registry_with_subtype_no_field():
    dtype_str = "cat[Record[Customer]]"
    result = parse_dtype(dtype_str)
    assert len(result) == 1
    assert result[0] == {
        "registry_str": "Record",
        "filter_str": "",
        "field_str": "name",
        "registry": Record,
        "field": Record.name,
        "subtypes_list": ["Customer"],
    }


def test_list_of_dtypes():
    dtype_str = "list[cat[Record[Customer]]]"
    result = parse_dtype(dtype_str)
    assert len(result) == 1
    assert result[0] == {
        "registry_str": "Record",
        "filter_str": "",
        "field_str": "name",
        "registry": Record,
        "field": Record.name,
        "subtypes_list": ["Customer"],
        "list": True,
    }
    assert serialize_dtype(list[bt.CellLine]) == "list[cat[bionty.CellLine]]"


def test_registry_with_filter():
    dtype_str = "cat[bionty.Gene.ensembl_gene_id[source__id='abcd']]"
    result = parse_dtype(dtype_str)
    assert len(result) == 1
    assert result[0] == {
        "registry_str": "bionty.Gene",
        "filter_str": "source__id='abcd'",
        "field_str": "ensembl_gene_id",
        "registry": bt.Gene,
        "field": bt.Gene.ensembl_gene_id,
        "subtypes_list": [],
    }


def test_nested_cat_dtypes():
    dtype_str = "cat[Record[Customer[UScustomer]].name]"
    result = parse_dtype(dtype_str)
    assert len(result) == 1
    assert result[0] == {
        "registry_str": "Record",
        "filter_str": "",
        "field_str": "name",
        "registry": Record,
        "field": Record.name,
        "subtypes_list": ["Customer", "UScustomer"],
    }


def test_nested_cat_with_filter():
    dtype_str = "cat[Record[Customer[UScustomer[region='US']]].description]"
    result = parse_dtype(dtype_str)
    assert len(result) == 1
    assert result[0] == {
        "registry_str": "Record",
        "filter_str": "region='US'",
        "field_str": "description",
        "registry": Record,
        "field": Record.description,
        "subtypes_list": ["Customer", "UScustomer"],
    }


# -----------------------------------------------------------------------------
# parsing django filter expressions
# -----------------------------------------------------------------------------


def test_feature_dtype():
    feature = ln.Feature(
        name="disease",
        dtype=bt.Disease,
        cat_filters={
            "source__uid": "4a3ejKuf"
        },  # uid corresponds to disease_ontology_old.uid
    ).save()

    result = parse_dtype(feature.dtype)
    assert len(result) == 1
    assert result[0] == {
        "registry_str": "bionty.Disease",
        "filter_str": "source__uid='4a3ejKuf'",
        "field_str": "name",
        "registry": bt.Disease,
        "field": bt.Disease.name,
        "subtypes_list": [],
    }

    feature.delete(permanent=True)


def test_cat_filters_incompatible_with_union_dtypes():
    with pytest.raises(ValidationError) as exc_info:
        ln.Feature(
            name="test_feature",
            dtype="cat[Record|bionty.CellType]",
            cat_filters={"source": "test"},
        )
    assert (
        "cat_filters are incompatible with union dtypes: 'cat[Record|bionty.CellType]'"
        in str(exc_info.value)
    )


def test_cat_filters_incompatible_with_nested_dtypes():
    with pytest.raises(ValidationError) as exc_info:
        ln.Feature(
            name="test_feature",
            dtype="cat[Record[Customer[SubCustomer]]]",
            cat_filters={"source": "test"},
        )
    assert (
        "cat_filters are incompatible with nested dtypes: 'cat[Record[Customer[SubCustomer]]]'"
        in str(exc_info.value)
    )


def test_parse_filter_string_basic():
    result = parse_filter_string("parent__id=123, category__name=electronics")
    expected = {
        "parent__id": ("parent", "id", "123"),
        "category__name": ("category", "name", "electronics"),
    }
    assert result == expected


def test_parse_filter_string_direct_fields():
    result = parse_filter_string("name=test, status=active")
    expected = {"name": ("name", None, "test"), "status": ("status", None, "active")}
    assert result == expected


def test_parse_filter_string_empty():
    with pytest.raises(ValueError) as e:
        parse_filter_string("")
        assert "missing '=' sign" in str(e)


def test_parse_filter_string_malformed():
    with pytest.raises(ValueError) as e:
        parse_filter_string("malformed_filter")
        assert "missing '=' sign" in str(e)


def test_parse_filter_string_missing_key():
    with pytest.raises(ValueError) as e:
        parse_filter_string("=someval")
        assert "empty key" in str(e)


def test_parse_filter_string_missing_value():
    with pytest.raises(ValueError) as e:
        parse_filter_string("somekey=")
        assert "empty val" in str(e)


def test_resolve_direct_fields():
    parsed = {"name": ("name", None, "test"), "status": ("status", None, "active")}
    result = resolve_relation_filters(parsed, bt.Gene)
    assert result == {"name": "test", "status": "active"}


def test_resolve_relation_filter_with_uid(source):
    parsed = {"source__uid": ("source", "uid", "testuid1")}
    result = resolve_relation_filters(parsed, bt.Gene)
    assert result == {"source": source}
    source.delete(permanent=True)


def test_resolve_relation_filter_with_name(organism):
    parsed = {"organism__name": ("organism", "name", "test_organism")}
    result = resolve_relation_filters(parsed, bt.Gene)
    assert result == {"organism": organism}
    organism.delete(permanent=True)


def test_resolve_multiple_relation_filters(organism, source):
    parsed = {
        "organism__name": ("organism", "name", "test_organism"),
        "source__uid": ("source", "uid", "testuid1"),
    }
    result = resolve_relation_filters(parsed, bt.Gene)
    assert result == {"organism": organism, "source": source}
    source.delete(permanent=True)
    organism.delete(permanent=True)


def test_resolve_nested_filter(organism):
    parsed = {"organism__name__contains": ("organism", "name__contains", "test_orga")}
    result = resolve_relation_filters(parsed, bt.Gene)
    assert result == {"organism": organism}
    organism.delete(permanent=True)


def test_resolve_relation_filter_failed_resolution():
    parsed = {"organism__name": ("organism", "name", "nonexistent")}
    with pytest.raises(bt.Organism.DoesNotExist):
        resolve_relation_filters(parsed, bt.Gene)


def test_resolve_relation_filter_duplicate():
    parsed = {
        "source__uid": ("source", "uid", "testuid1"),
        "source__name": ("source", "name", "test_name"),
    }
    with pytest.raises(bt.Source.DoesNotExist):
        resolve_relation_filters(parsed, bt.Gene)
